### HERE WE DEFINE A SET OF FUNCTIONS THAT COMPUTE CODON USAGE BIAS
### INDICATOR MEASURES FOR DIFFERENT MRNA SEQUENCES OF SACCHAROMYCES CERIVISAE S288C.
### THIS IS ALSO PERFORMED FOR SEQUENCES THAT ARE INTERRUPTED BY PROGRAMMED RIBOSOMAL FRAMESHIFTING (-1 SHIFT).
### AND STOP (MOST OF THE TIME) AT A PREMATURE STOP CODON (PSC).
### NAMELY THE GHAEMMAGHANI DATASET (GHAEMMAGHANI ET AL., 2003, NATURE), WHICH
### THESE COMPUTATIONS ARE PERFORMED ON TWO DATASETS THAT HAVE BEEN CONVERTED INTO A DATABASE FROM THE
### MARYLAND UNIVERSITY PRF-1 DATASET.
###
### Victor Garcia, Zurich University of Applied Sciences (ZHAW), May 2017
########################################################################################################################

import pymysql.cursors
import json
import matplotlib.pyplot
#import Bio
import numpy
from scipy import stats

### Find information: http://biopython.org/wiki/SeqIO
from Bio import SeqIO



#import pymysql


# Connect to the database
#connection = pymysql.connect(host= '127.0.0.1',
#                             user='victor1',
#                             passwd='',
#                             db='test',
#                             cursorclass=pymysql.cursors.DictCursor)



# Connect to the database
connection = pymysql.connect(host='127.0.0.1',
                             user='root',
                             passwd='MinkowskiaN1012',
                             db='prfdb_mysql',
                             cursorclass=pymysql.cursors.DictCursor)


#### this thing creates an object that allows to run queries on the database
cursor = connection.cursor()


#sql = 'SELECT * FROM mfe_S288C  WHERE mfe_method = "hotknots" '

### this is the name of one of the methods used to identify slippery sites in the database
method = "nupack"

### this is a string for a query. One needs to specify the method.
sql = 'SELECT accession, start, slipsite, sequence FROM mfe_S288C WHERE mfe_method = %s'

### running the query. The variable status is going to show whether the execution ran successfully (or something like that)
status = cursor.execute(sql, method)

### here we store the actual data from the query, which is stored in the cursor object, under the object rows
slip_data = cursor._rows

### test whether it worked
print slip_data[0]


### this is a string for a query. One needs to specify the method.
sql_wholegenome = 'SELECT accession, genename, mrna_seq FROM genome'

status = cursor.execute(sql_wholegenome)
whole_data = cursor._rows

### test whether it worked

print whole_data[0]
print len(whole_data)

### every entry of data is a dictionary. The first entry is the the name of the gene.

#for row in cursor:
#  print(row)

########################################################################################################################
################ compute the nucleotide frequencies of inter-genic sequences ################

#from Bio import AlignIO
# record = SeqIO.read("NotFeature.fasta", "fasta")
# record = SeqIO.read(open("NotFeature.fasta"), "fasta")
# record = AlignIO.read("NotFeature.fasta", "fasta")
# record = AlignIO.read(open("NotFeature.fasta"), "fasta")
#for record in SeqIO.parse("NotFeature.fasta", "fasta"):
#    print(record.id)

### this one works
records = list(SeqIO.parse("NotFeature.fasta", "fasta"))
#records[1].seq
#records[1].seq.__seq__()
#records[1].seq.tostring()

### how to access the string in the record
# test_seq = str(records[1].seq)
# test_seq.replace("T", "U")
# codon_count(test_seq.replace("T", "U"))


### concatenate

### some info here: http://biopython.org/wiki/SeqRecord
#for index, record in enumerate(SeqIO.parse("NotFeature.fasta", "fasta")):
#    print("index %i, ID = %s, length %i, with %i features" % (index, record.id, len(record.seq), len(record.features)))




### all of the strings concatenated
all_str = ''

### loop through records object, and add the strings together
for i in range(0,len(records)):

    ### retrieve string
    m_str = str(records[i].seq)
    ### retrieve string
    m_str = m_str.replace("T", "U")

    ### concatenate
    all_str += m_str


################


def nucleotide_freqs(seq):

    ### set up frequency of nucleotide variables
    p_A = 0.0
    p_U = 0.0
    p_C = 0.0
    p_G = 0.0
    tot_count = len(seq)

    ### loop through the sequence
    for nucl in seq:
        if nucl == 'A':
            p_A += 1.0
        if nucl == 'U':
            p_U += 1.0
        if nucl == 'C':
            p_C += 1.0
        if nucl == 'G':
            p_G += 1.0

    ### normalize
    p_A = p_A / tot_count
    p_U = p_U / tot_count
    p_C = p_C / tot_count
    p_G = p_G / tot_count

    ### create dictionary
    freqs = {'A': p_A, 'U': p_U, 'C': p_C, 'G': p_G}

    return freqs

### test
# nucleotide_freqs(test_seq.replace("T", "U"))


### dictionary of nucleotide frequencies needed later
nucl_freqs = nucleotide_freqs(all_str)

########################################################################################################################



########################################################################################################################
################ FUNCTIONS NEEDED FOR COMPUTATIONS ################
########################################################################################################################


### define genetic code
def genetic_code(codon):

    ### returns an amino acid depending on the codon
    if codon == "UAG" or codon == "UAA" or codon == "UGA":
        return "stop"

    elif codon ==  "AUU" or codon == "AUC" or codon == "AUA":
        return "isoleucine" ## "I"

    elif codon ==  "CUU" or codon == "CUC" or codon == "CUA" or codon == "CUG" or codon == "UUA" or codon == "UUG":
        return "leucine" ## "L"

    elif codon ==  "GUU" or codon == "GUC" or codon == "GUA" or codon == "GUG":
        return "valine" ## "V"

    elif codon == "UUU" or codon == "UUC":
        return "phenylalanine" ## "F"

    elif codon == "AUG":
        return "methionine" ## "M"

    elif codon == "UGU" or codon == "UGC":
        return "cysteine" ## "C"

    elif codon ==  "GCU" or codon == "GCC" or codon == "GCA" or codon == "GCG":
        return "alanine" ## "A"

    elif codon == "GGU" or codon == "GGC" or codon == "GGA" or codon == "GGG":
        return "glycine" ## "G"

    elif codon == "CCU" or codon == "CCC" or codon == "CCA" or codon == "CCG":
        return "proline" ## "P"

    elif codon == "ACU" or codon == "ACC" or codon == "ACA" or codon == "ACG":
        return "threonine" ## "U"

    elif codon == "UCU" or codon == "UCC" or codon == "UCA" or codon == "UCG" or codon == "AGU" or codon == "AGC":
        return "serine" ## "S"

    elif codon == "UAU" or codon == "UAC":
        return "tyrosine" ## "Y"

    elif codon == "UGG":
        return "tryptophan" ## "W"

    elif codon == "CAA" or codon == "CAG":
        return "glutamine" ## "Q"

    elif codon ==  "AAU" or codon == "AAC":
        return "asparagine" ## "N"

    elif codon == "CAU" or codon == "CAC":
        return "histidine" ## "H"

    elif codon == "GAA" or codon == "GAG":
        return "glutamic acid" ## "E"

    elif codon ==  "GAU" or codon == "GAC":
        return "aspartic acid" ## "D"

    elif codon ==  "AAA" or codon == "AAG":
        return "lysine" ## "K"

    elif codon == "CGU" or codon == "CGC" or codon == "CGA" or codon == "CGG" or codon == "AGA" or codon == "AGG":
        return "arginine" ## "R"
    else:
        return "not found"


### returns the codons that decode for an amino acid
def codons_in_amino(amino_key):

    ### define a library of the codons decoding for each amino acid
    amino_codon_library = {"stop": {"UAG": 0, "UAA": 0, "UGA": 0},
                           "isoleucine": {"AUU": 0, "AUC": 0, "AUA": 0},
                           "leucine": {"CUU": 0, "CUC": 0, "CUA": 0, "CUG": 0, "UUA": 0, "UUG": 0},
                           "valine": {"GUU": 0, "GUC": 0, "GUA": 0, "GUG": 0},
                           "phenylalanine": {"UUU": 0, "UUC": 0},
                           "methionine": {"AUG": 0},
                           "cysteine": {"UGU": 0, "UGC": 0},
                           "alanine": {"GCU": 0, "GCC": 0, "GCA": 0, "GCG": 0},
                           "glycine": {"GGU": 0, "GGC": 0, "GGA": 0, "GGG": 0},
                           "proline": {"CCU": 0, "CCC": 0, "CCA": 0, "CCG": 0},
                           "threonine": {"ACU": 0, "ACC": 0, "ACA": 0, "ACG": 0},
                           "serine": {"UCU": 0, "UCC": 0, "UCA": 0, "UCG": 0, "AGU": 0, "AGC": 0},
                           "tyrosine": {"UAU": 0, "UAC": 0},
                           "tryptophan": {"UGG": 0},
                           "glutamine": {"CAA": 0, "CAG": 0},
                           "asparagine": {"AAU": 0, "AAC": 0},
                           "histidine": {"CAU": 0, "CAC": 0},
                           "glutamic acid": {"GAA": 0, "GAG": 0},
                           "aspartic acid": {"GAU": 0, "GAC": 0},
                           "lysine": {"AAA": 0, "AAG": 0},
                           "arginine": {"CGU": 0, "CGC": 0, "CGA": 0, "CGG": 0, "AGA": 0, "AGG": 0}}


    cod_dict = amino_codon_library[amino_key]

    return cod_dict


### test
# codons_in_amino("isoleucine")


### returns the codons that decode for an amino acid
def degen_amino(amino_key):

    ### define a library of the codons decoding for each amino acid
    amino_degen_library = {"stop": 3,
                           "isoleucine": 3,
                           "leucine": 6,
                           "valine": 4,
                           "phenylalanine": 2,
                           "methionine": 1,
                           "cysteine": 2,
                           "alanine": 4,
                           "glycine": 4,
                           "proline": 4,
                           "threonine": 4,
                           "serine": 6,
                           "tyrosine": 2,
                           "tryptophan": 1,
                           "glutamine": 2,
                           "asparagine": 2,
                           "histidine": 2,
                           "glutamic acid": 2,
                           "aspartic acid": 2,
                           "lysine": 2,
                           "arginine": 6}


    cod_degen = amino_degen_library[amino_key]

    return cod_degen


### test
#degen_amino("histidine")

### counts the number of each codon in an mRNA sequence
def codon_count(sequence, finish_with_stop = False):

    ### dictionary with all codons, and their respective counts
    cod_count = {"UAG": 0, "UAA": 0, "UGA": 0, "AUU": 0, "AUC": 0, "AUA": 0, "CUU": 0, "CUC": 0, "CUA": 0, "CUG": 0, "UUA": 0, "UUG": 0,
                 "GUU": 0, "GUC": 0, "GUA": 0, "GUG": 0, "UUU": 0, "UUC": 0, "AUG": 0, "UGU": 0, "UGC": 0, "GCU": 0, "GCC": 0, "GCA": 0,
                 "GCG": 0, "GGU": 0, "GGC": 0, "GGA": 0, "GGG": 0, "CCU": 0, "CCC": 0, "CCA": 0, "CCG": 0, "ACU": 0, "ACC": 0, "ACA": 0,
                 "ACG": 0, "UCU": 0, "UCC": 0, "UCA": 0, "UCG": 0, "AGU": 0, "AGC": 0, "UAU": 0, "UAC": 0, "UGG": 0, "CAA": 0, "CAG": 0,
                 "AAU": 0, "AAC": 0, "CAU": 0, "CAC": 0, "GAA": 0, "GAG": 0, "GAU": 0, "GAC": 0, "AAA": 0, "AAG": 0, "CGU": 0, "CGC": 0,
                 "CGA": 0, "CGG": 0, "AGA": 0, "AGG": 0}

    ### go through sequence and add count
    seq_length = len(sequence)
    ### seq_length should be divisible by three
    L = seq_length/3

    ### what do we do if it is not? In fact, often times it is not divisible by three.

    for k in range(0, L, 1):
        cod = sequence[3*k:3*k + 3]
        if genetic_code(cod) == "stop" and finish_with_stop:
            print "stop codon found at site %d" % float(3*k)
            break
        cod_count[cod] += 1

    return cod_count

### test
# test_seq = slip_data[1]["sequence"]
# codon_count(test_seq)

### function to print the codon counts on the console
def print_codcount(cod_count):
    for key in cod_count:
        print key + " = " + str(cod_count[key])


#### test
# test_cod_count = codon_count(slip_data[1]["sequence"])
# print_codcount(test_cod_count)


### calculates the frequency of each codon in an mRNA sequence (not within amino acid)
def codon_frequency(sequence, finish_with_stop = False):

    ### dictionary with all codons, and their respective counts
    cod_count = {"UAG": 0.0, "UAA": 0.0, "UGA": 0.0, "AUU": 0.0, "AUC": 0.0, "AUA": 0.0, "CUU": 0.0, "CUC": 0.0, "CUA": 0.0, "CUG": 0.0,
                 "UUA": 0.0, "UUG": 0.0,
                 "GUU": 0.0, "GUC": 0.0, "GUA": 0.0, "GUG": 0.0, "UUU": 0.0, "UUC": 0.0, "AUG": 0.0, "UGU": 0.0, "UGC": 0.0, "GCU": 0.0,
                 "GCC": 0.0, "GCA": 0.0,
                 "GCG": 0.0, "GGU": 0.0, "GGC": 0.0, "GGA": 0.0, "GGG": 0.0, "CCU": 0.0, "CCC": 0.0, "CCA": 0.0, "CCG": 0.0, "ACU": 0.0,
                 "ACC": 0.0, "ACA": 0.0,
                 "ACG": 0.0, "UCU": 0.0, "UCC": 0.0, "UCA": 0.0, "UCG": 0.0, "AGU": 0.0, "AGC": 0.0, "UAU": 0.0, "UAC": 0.0, "UGG": 0.0,
                 "CAA": 0.0, "CAG": 0.0,
                 "AAU": 0.0, "AAC": 0.0, "CAU": 0.0, "CAC": 0.0, "GAA": 0.0, "GAG": 0.0, "GAU": 0.0, "GAC": 0.0, "AAA": 0.0, "AAG": 0.0,
                 "CGU": 0.0, "CGC": 0.0,
                 "CGA": 0.0, "CGG": 0.0, "AGA": 0.0, "AGG": 0.0}

    ### go through sequence and add count
    seq_length = len(sequence)
    ### seq_length should be divisible by three
    L = seq_length / 3

    ### what do we do if it is not? In fact, often times it is not divisible by three.
    tot_count = 0.0


    for k in range(0, L, 1):
        cod = sequence[3 * k:3 * k + 3]
        if genetic_code(cod) == "stop" and finish_with_stop:
            print "stop codon found at site %d" % float(3 * k)
            break
        cod_count[cod] += 1
        tot_count += 1.0

    ### normalize everything in the dictionary by tot_count
    cod_freq = {k: v / tot_count for k, v in cod_count.iteritems()}

    return cod_freq

### test
# test_seq = slip_data[1]["sequence"]
# test_seq = whole_data[1]["mrna_seq"].replace("T", "U")
# codon_frequency(test_seq)




### compute the relative adaptiveness values
### of the codons found in a sequence
def rel_adaptiveness(sequence):


    ### first compute the counts of each codon
    cod_count = codon_count(sequence)

    ### copy that into a relative adaptiveness object
    rel_adap = cod_count

    ### for debugging
    #print_codcount(cod_count)

    ### need to compute the maximum codon count
    ### among all codons decoding for the same amino acid
    ### this is an amino-acid specific dictionary
    ### define a library of the codons decoding for each amino acid
    amino_codon_library = {"stop": {"UAG": 0, "UAA": 0, "UGA": 0},
                           "isoleucine": {"AUU": 0, "AUC": 0, "AUA": 0},
                           "leucine": {"CUU": 0, "CUC": 0, "CUA": 0, "CUG": 0, "UUA": 0, "UUG": 0},
                           "valine": {"GUU": 0, "GUC": 0, "GUA": 0, "GUG": 0},
                           "phenylalanine": {"UUU": 0, "UUC": 0},
                           "methionine": {"AUG": 0},
                           "cysteine": {"UGU": 0, "UGC": 0},
                           "alanine": {"GCU": 0, "GCC": 0, "GCA": 0, "GCG": 0},
                           "glycine": {"GGU": 0, "GGC": 0, "GGA": 0, "GGG": 0},
                           "proline": {"CCU": 0, "CCC": 0, "CCA": 0, "CCG": 0},
                           "threonine": {"ACU": 0, "ACC": 0, "ACA": 0, "ACG": 0},
                           "serine": {"UCU": 0, "UCC": 0, "UCA": 0, "UCG": 0, "AGU": 0, "AGC": 0},
                           "tyrosine": {"UAU": 0, "UAC": 0},
                           "tryptophan": {"UGG": 0},
                           "glutamine": {"CAA": 0, "CAG": 0},
                           "asparagine": {"AAU": 0, "AAC": 0},
                           "histidine": {"CAU": 0, "CAC": 0},
                           "glutamic acid": {"GAA": 0, "GAG": 0},
                           "aspartic acid": {"GAU": 0, "GAC": 0},
                           "lysine": {"AAA": 0, "AAG": 0},
                           "arginine": {"CGU": 0, "CGC": 0, "CGA": 0, "CGG": 0, "AGA": 0, "AGG": 0}}


    ### go through sequence and add count
    seq_length = len(sequence)

    ### seq_length should be divisible by three (number of codons)
    L = seq_length/3

    ### L will be an integer

    ### loop through all amino acids. Fill the amino_codon_library
    for k in range(0, L, 1):

        ### extract the relevant codon from the sequence
        cod = sequence[(3*k):(3*k + 3)]

        ### if that is a stop codon, stop the procedure
        #if genetic_code(cod) == "stop":
        #    print "stop codon found"
        #    break


        amino_codon_library[genetic_code(cod)][cod] += 1


    for cod in rel_adap:

        amino = genetic_code(cod)
        max_cod = max(amino_codon_library[amino], key = lambda i: amino_codon_library[amino][i])

        rel_adap[cod] = float(rel_adap[cod])/amino_codon_library[amino][max_cod]

    return rel_adap


### test
# test_seq = whole_data[1]["mrna_seq"].replace("T", "U")
# rel_adaptiveness(test_seq)

### retrieves the codon counts from all the ribosomal genes of yeast:
### For each ribosomal gene sequence, a codon count dictionary is computed,
### and then added together tro one unyfing count.
def retrieve_ribo_seq_count(whole_data):

    ribo_genes_smallsubunit = ["RPS0A", "RPS0B", "RPS2", "RPS3", "RPS1A", "RPS1B", "RPS4A", "RPS4B",
                               "RPS5", "RPS6A", "RPS6B", "RPS7A", "RPS7B", "RPS8A", "RPS8B", "RPS9A",
                               "RPS9B", "RPS10A", "RPS10B", "RPS11A", "RPS11B", "RPS12", "RPS13",
                               "RPS14A", "RPS14B", "RPS15", "RPS22A", "RPS22B", "RPS16A", "RPS16B",
                               "RPS17A", "RPS17B", "RPS18A", "RPS18B", "RPS19A", "RPS19B", "RPS20",
                               "RPS21A", "RPS21B", "RPS23A", "RPS23B", "RPS24A", "RPS24B", "RPS25A",
                               "RPS25B", "RPS26A",  "RPS26B", "RPS27A", "RPS27B", "RPS31", "RPS28A",
                               "RPS28B", "RPS29A", "RPS29B", "RPS30A", "RPS30B"]


    ribo_genes_largesubunit = ["RPL3", "RPL4A", "RPL4B", "RPL5", "RPL6A", "RPL6B", "RPL7A", "RPL7B", "RPL8A",
                               "RPL8B", "RPL2A", "RPL2B", "RPL9A", "RPL9B", "RPL10", "RPL1A", "RPL1B", "RPL11A",
                               "RPL11B", "RPL12A", "RPL12B", "RPL13A", "RPL13B", "RPL16A", "RPL16B", "RPL14A",
                               "RPL14B", "RPL15A", "RPL15B", "RPL17A", "RPL17B", "RPL18A", "RPL18B", "RPL20A",
                               "RPL20B", "RPL19A", "RPL19B", "RPL21A", "RPL21B", "RPL22A", "RPL22B", "RPL23A",
                               "RPL23B", "RPL25", "RPL24A", "RPL24B", "RPL26A", "RPL26B", "RPL27A", "RPL27B",
                               "RPL28", "RPL29", "RPL30", "RPL31A", "RPL31B", "RPL32", "RPL34A", "RPL34B",
                               "RPL35A", "RPL35B", "RPL33A", "RPL33B", "RPL36A", "RPL36B", "RPL42A",
                               "RPL42B", "RPL37A", "RPL37B", "RPL43A", "RPL43B", "RPL38", "RPL39",
                               "RPL40A", "RPL40B", "RPL41A", "RPL41B", "RPP0", "RPP1A", "RPP1B", "RPP2A"]


    tot_cod_count = {}

    for i in range(0,len(ribo_genes_smallsubunit)):

        gene_ind = retrieve_wd_geneind(whole_data, ribo_genes_smallsubunit[i])

        ribo_seq = whole_data[gene_ind]["mrna_seq"]

        ribo_cod_count = codon_count(ribo_seq)

        tot_cod_count = combine_cod_counts(tot_cod_count, ribo_cod_count)


    for i in range(0,len(ribo_genes_largesubunit)):

        gene_ind = retrieve_wd_geneind(whole_data, ribo_genes_largesubunit[i])

        ribo_seq = whole_data[gene_ind]["mrna_seq"]

        ribo_cod_count = codon_count(ribo_seq)

        tot_cod_count = combine_cod_counts(tot_cod_count, ribo_cod_count)


    return tot_cod_count





### combines the codon counts of two different sequences together
### by adding them, and returns that sum
def combine_cod_counts(cod_count_A, cod_count_B):

    combi_cod_conti = {}

    for i in cod_count_A:

        print i
        combi_cod_conti[i] = cod_count_A[i] + cod_count_B[i]

    return combi_cod_conti



### retrieves a concatenation of ribosomal mrna
def retrieve_ribo_seq(whole_data):

    ribo_genes_smallsubunit = ["RPS0A", "RPS0B", "RPS2", "RPS3", "RPS1A", "RPS1B", "RPS4A", "RPS4B",
                               "RPS5", "RPS6A", "RPS6B", "RPS7A", "RPS7B", "RPS8A", "RPS8B", "RPS9A",
                               "RPS9B", "RPS10A", "RPS10B", "RPS11A", "RPS11B", "RPS12", "RPS13",
                               "RPS14A", "RPS14B", "RPS15", "RPS22A", "RPS22B", "RPS16A", "RPS16B",
                               "RPS17A", "RPS17B", "RPS18A", "RPS18B", "RPS19A", "RPS19B", "RPS20",
                               "RPS21A", "RPS21B", "RPS23A", "RPS23B", "RPS24A", "RPS24B", "RPS25A",
                               "RPS25B", "RPS26A",  "RPS26B", "RPS27A", "RPS27B", "RPS31", "RPS28A",
                               "RPS28B", "RPS29A", "RPS29B", "RPS30A", "RPS30B"]


    ribo_genes_largesubunit = ["RPL3", "RPL4A", "RPL4B", "RPL5", "RPL6A", "RPL6B", "RPL7A", "RPL7B", "RPL8A",
                               "RPL8B", "RPL2A", "RPL2B", "RPL9A", "RPL9B", "RPL10", "RPL1A", "RPL1B", "RPL11A",
                               "RPL11B", "RPL12A", "RPL12B", "RPL13A", "RPL13B", "RPL16A", "RPL16B", "RPL14A",
                               "RPL14B", "RPL15A", "RPL15B", "RPL17A", "RPL17B", "RPL18A", "RPL18B", "RPL20A",
                               "RPL20B", "RPL19A", "RPL19B", "RPL21A", "RPL21B", "RPL22A", "RPL22B", "RPL23A",
                               "RPL23B", "RPL25", "RPL24A", "RPL24B", "RPL26A", "RPL26B", "RPL27A", "RPL27B",
                               "RPL28", "RPL29", "RPL30", "RPL31A", "RPL31B", "RPL32", "RPL34A", "RPL34B",
                               "RPL35A", "RPL35B", "RPL33A", "RPL33B", "RPL36A", "RPL36B", "RPL42A",
                               "RPL42B", "RPL37A", "RPL37B", "RPL43A", "RPL43B", "RPL38", "RPL39",
                               "RPL40A", "RPL40B", "RPL41A", "RPL41B", "RPP0", "RPP1A", "RPP1B", "RPP2A"]


    tot_ribo_seq = ""

    for i in range(0,len(ribo_genes_smallsubunit)):

        gene_ind = retrieve_wd_geneind(whole_data, ribo_genes_smallsubunit[i])

        ribo_seq = whole_data[gene_ind]["mrna_seq"]

        tot_ribo_seq = tot_ribo_seq + ribo_seq


    for i in range(0,len(ribo_genes_largesubunit)):

        gene_ind = retrieve_wd_geneind(whole_data, ribo_genes_largesubunit[i])

        ribo_seq = whole_data[gene_ind]["mrna_seq"]

        tot_ribo_seq = tot_ribo_seq + ribo_seq


    return tot_ribo_seq

### test
# ribo_seq_test = retrieve_ribo_seq(whole_data)
# ribo_seq_test = ribo_seq_test.replace("T", "U")

### search for the stop codon in the shifted frame
def CAI(sequence, ribosome_seq):

    ### relative adaptiveness measures as
    ### described in Roth, Anisimova and Cannarozzi,
    ### Codon evolution: mechanisms and models.
    ### New York: Oxford University Press Inc (2012): 189-217.
    w = rel_adaptiveness(ribosome_seq)

    ### go through sequence and add count
    seq_length = len(sequence)

    ### seq_length should be divisible by three (number of codons)
    L = seq_length/3

    ### count the instances of each codon
    ### cod_count = o_ac in Roth, Anisimova, Cannarozzi 2012 ("Measuring Codon Usage Bias")
    cod_count = codon_count(sequence)

    ### define cai val
    cai = 0.0

    ### loop through all codons and form the sum of the
    ### observed count times the log of the relative adaptiveness of the reference sequence
    for i in range(0, len(cod_count)):

        cod_key = cod_count.keys()[i]

        #print cod_key + " with count: " + str(cod_count[cod_key]) + " has rel value: " + str(w[cod_key])

        if w[cod_key] > 0.0:
            cai += cod_count[cod_key]*numpy.log(w[cod_key])

    cai = cai/L
    cai = numpy.exp(cai)

    return cai




### search for the stop codon in the shifted frame
### ribo_rel_adap = w_ac in Roth, Anisimova, Cannarozzi 2012 ("Measuring Codon Usage Bias")
def CAI_f(sequence, ribo_rel_adap):

    ### relative adaptiveness measures as
    ### described in Roth, Anisimova and Cannarozzi,
    ### Codon evolution: mechanisms and models.
    ### New York: Oxford University Press Inc (2012): 189-217.
    w = ribo_rel_adap

    ### go through sequence and add count
    seq_length = len(sequence)

    ### seq_length should be divisible by three (number of codons)
    L = seq_length/3

    ### count the instances of each codon
    ### cod_count = o_ac in Roth, Anisimova, Cannarozzi 2012 ("Measuring Codon Usage Bias")
    cod_count = codon_count(sequence)

    ### define cai val
    cai = 0.0

    for i in range(0, len(cod_count)):

        cod_key = cod_count.keys()[i]

        #print cod_key + " with count: " + str(cod_count[cod_key]) + " has rel value: " + str(w[cod_key])

        if w[cod_key] > 0.0:
            cai += cod_count[cod_key]*numpy.log(w[cod_key])

    cai = cai/L
    cai = numpy.exp(cai)

    return cai



### computation of the codon usabe index B (Karlin and Mrazek 1996)
### seq_cod_freq = f_ac, ref_cod_freq = f^ref_ac, ref_amino_freq = F_a
def B(seq_cod_freq, ref_cod_freq, ref_amino_freq):

    ### initialize cbi
    b = 0.0


    for amin_key in ref_amino_freq:

        ### Compute F_a
        F_a = ref_amino_freq[amin_key]

        ### initialize d_a
        d_a = 0.0

        ### compute d_a
        for cod_key in codons_in_amino(amin_key):

            d_a += abs(ref_cod_freq[cod_key] - seq_cod_freq[cod_key])

        ### add to the
        b += F_a*d_a

    return b


### compute expected codon frequency within
### a codon familiy decoding for the same amino acid
def expec_cod_freq(codon, nucl_freqs):

    expec_prob = 1.0

    for n in range(0,3):
        expec_prob = expec_prob*nucl_freqs[codon[n]]

    prob_codons_of_amino = codons_in_amino(genetic_code(codon))
    degen_amino = len(prob_codons_of_amino)
    for cods in prob_codons_of_amino:

        expec_prob_amino = 1.0

        for n in range(0, 3):
            expec_prob_amino = expec_prob_amino*nucl_freqs[cods[n]]

            prob_codons_of_amino[cods] = expec_prob_amino


    ### the sum of the probabilities within the same amino acid family
    prob_sum = sum(prob_codons_of_amino.values())
    expec_prob = expec_prob/prob_sum

    return expec_prob

### test
#expec_cod_freq('CAU', nucl_freqs)
#expec_cod_freq('CAC', nucl_freqs)


### computes N_c^prime by JA Novembre, 2002, smbe
def Ncprime(seq, n_freqs, verbose = False):

    ### normalized codon frequencies
    ### (normalized within same codon family decoding for the same amino acid)
    c_freq = codon_freq_in_amino(seq)
    #expec_cod_freq(seq, n_freqs)
    am_count = amino_count(codon_count(seq))

    ### define degeneracy classes
    degen_class = {"onefold": ["methionine", "tryptophan"],
                   "twofold": ["phenylalanine", "cysteine", "tyrosine", "glutamine", "asparagine", "histidine", "glutamic acid", "aspartic acid", "lysine"],
                   "threefold": ["isoleucine"],
                   "fourfold": ["valine", "alanine", "glycine", "proline", "threonine"],
                   "sixfold": ["leucine", "serine", "arginine"]}


    ### counts of codons within an amino acid
    n_a = {"isoleucine": 0.0, "leucine": 0.0, "valine": 0.0, "phenylalanine": 0.0, "methionine": 0.0,
                    "cysteine": 0.0, "alanine": 0.0, "glycine": 0.0, "proline": 0.0, "threonine": 0.0, "serine": 0.0,
                    "tyrosine": 0.0, "tryptophan": 0.0, "glutamine": 0.0, "asparagine": 0.0, "histidine": 0.0,
                    "glutamic acid": 0.0, "aspartic acid": 0.0, "lysine": 0.0, "arginine": 0.0}

    ### homozygosity
    F_a = {"isoleucine": 0.0, "leucine": 0.0, "valine": 0.0, "phenylalanine": 0.0, "methionine": 0.0,
                    "cysteine": 0.0, "alanine": 0.0, "glycine": 0.0, "proline": 0.0, "threonine": 0.0, "serine": 0.0,
                    "tyrosine": 0.0, "tryptophan": 0.0, "glutamine": 0.0, "asparagine": 0.0, "histidine": 0.0,
                    "glutamic acid": 0.0, "aspartic acid": 0.0, "lysine": 0.0, "arginine": 0.0}


    ### define degeneracy classes
    F_d = {"onefold": 0.0, "twofold": 0.0, "threefold": 0.0, "fourfold": 0.0, "sixfold": 0.0}

    Nc = 0.0

    ### loop through degeneracy classes
    for d in degen_class:

        aminos_in_degen_class = degen_class[d]
        non_zero_amino_counts = 0.0

        if verbose:
            print "START OF A DEGENERACY CLASS"
            print "degeneracy class is: " + str(d)

        for am_ac in aminos_in_degen_class:

            ### chi for amino acid squared
            Chi_am_squared = 0.0
            n_a[am_ac] = am_count[am_ac] #cod_count[cod]

            if verbose:
                print "amino acid is: " + str(am_ac)
                print "na_[" + str(am_ac) + "] " + str(n_a[am_ac])

            cod_list = codons_in_amino(am_ac)

            for cod in cod_list:
                if verbose:
                    print "codon is %s, expec_cod_freq: %s" % (cod, str(expec_cod_freq(cod,n_freqs)))
                    print "observed frequency: %s" % str(c_freq[cod])
                #print "expec_codon_freq: %d" % expec_cod_freq(cod,n_freqs)
                expec_cod_f = expec_cod_freq(cod,n_freqs) ### expected codon frequency
                observed_cod_f = c_freq[cod] ### observed codon frequency (normalized)
                Chi_am_squared += ((observed_cod_f - expec_cod_f)**2)/expec_cod_f

            Chi_am_squared *= n_a[am_ac]

            if verbose:
                print "Chi**2[" + str(am_ac) + "]: " + str(Chi_am_squared)

            ### testing
            #for cod in cod_list:
            #    print c_freq[cod]

            ### testing
            #for cod in cod_list:
            #    print expec_cod_freq(cod,n_freqs)


            ### Novembre: "care must be taken to exclude F_a values that
            ### are undefined (n_a -1 = 0) or equal to zero."
            if n_a[am_ac] > 1.0: ### means F_a is defined or reasonable
                F_a[am_ac] = Chi_am_squared + n_a[am_ac] - float(degen_amino(am_ac))
                F_a[am_ac] = F_a[am_ac] / (float(degen_amino(am_ac)) * (n_a[am_ac] - 1.0))
            else:
                F_a[am_ac] = 0.0


            if verbose:
                print "F_a[" + str(am_ac) + "]: " + str(F_a[am_ac])

            ###  "or equal to zero."
            if F_a[am_ac] != 0.0: ### means that F_a is not zero, and is therefore not excluded (added to F_d[d]
                F_d[d] += F_a[am_ac]
                non_zero_amino_counts += 1.0

            if verbose:
                print "number of different observed amino_acids in degeneracy class " + str(d) + " is: " + str(non_zero_amino_counts)


        if verbose:
            print "averaging homozygosity values across degeneracy class"


        ### perform averaging
        if non_zero_amino_counts > 0.0:
            ### average F_d by all the amino acids within same degeneracy class
            F_d[d] = F_d[d] / non_zero_amino_counts #float(len(degen_class[d]))
        else:
            F_d[d] = 0.0

        if verbose:
            print "averaged F_d[" + str(d) + "] is " + str(F_d[d])


        ### Novembre: "if k-fold redundancy classes are unobserved, F_d[d] is assumed to be 1/k (k is number of amino acids of same degeneracy class)"
        if F_d[d] != 0:

            if verbose:
                print "F_d[" + str(d) + "] non-zero. Proceeding to add to  Nc\' after averaging"

            Nc += float(len(degen_class[d])) / F_d[d]

        elif F_d["threefold"] == 0.0: ### Novembre: "if F_3["threefold"] is unobserved, it should be estimated by averaging F_2 and F_4.

            if verbose:
                print "Threefold degenerate class is zero. No addition to Nc\'."

            Nc += 0.0

        elif F_d[d] == 0.0: ### this will not be executed if first looped through "threefold"

            if verbose:
                print "One degeneracy class is empty. "

            ### This is ugly code: correct it! float(degen_amino(am_ac)) is onl defined in inner for loop above
            F_d[d] = 1/float(degen_amino(am_ac))
            Nc += float(len(degen_class[d])) / F_d[d]

        if verbose:

            print "F_d[d] is: " + str(F_d[d])
            print "natural number of amino acids in this degen class is: " + str(len(degen_class[d]))
            print "number of different observed amino_acids in degeneracy class " + str(d) + " is: " + str(non_zero_amino_counts)
            print "addition to Nc\' is: " + str(float(len(degen_class[d])) / F_d[d])
            print "END OF A DEGENERACY CLASS"

    ### Novembre: "if F_3["threefold"] is unobserved, it should be estimated by averaging F_2 and F_4.
    if F_d["threefold"] == 0.0 and F_d["twofold"] > 0.0 and F_d["fourfold"] > 0.0 and F_d["sixfold"] > 0.0:

        if verbose:
            print "averaging procedure for threefold degeneracy"

        ### As suggested by Wright:
        #F_three_avg = (F_d["twofold"] + F_d["fourfold"])/2.0
        ### As suggested by Fulsang:
        F_three_avg = 1/3.0*((2/F_d["twofold"] -1)**(-1.0) + (2/(3.0*F_d["fourfold"]) - 1.0/3.0)**(-1.0) + (2/(5.0*F_d["sixfold"]) - 3.0/5.0)**(-1.0) )
        Nc += len(degen_class["threefold"]) / F_three_avg

    ### if nothing works, simply replace by "crude" frequency of usage of codons
    if Nc == 0.0:
        ### replace simply by number of used codons
        #sum([1 for x in c_freq if c_freq[x] > 0]) does not seem to count the stop codons as well
        Nc = sum([1 for x in c_freq if c_freq[x] > 0])


    return Nc

### test
#test_seq = whole_data[1]["mrna_seq"].replace("T", "U")
#Ncprime(test_seq, nucl_freqs)
#ribo_seq_test = retrieve_ribo_seq(whole_data)
#ribo_seq_test = ribo_seq_test.replace("T", "U")
#Ncprime(ribo_seq_test, nucl_freqs)


### computes N_c by Wright
def Nc(seq, verbose = False):

    ### normalized codon frequencies
    ### (normalized within same codon family decoding for the same amino acid)
    c_freq = codon_freq_in_amino(seq)
    #expec_cod_freq(seq, n_freqs)
    am_count = amino_count(codon_count(seq))

    ### define degeneracy classes
    degen_class = {"onefold": ["methionine", "tryptophan"],
                   "twofold": ["phenylalanine", "cysteine", "tyrosine", "glutamine", "asparagine", "histidine", "glutamic acid", "aspartic acid", "lysine"],
                   "threefold": ["isoleucine"],
                   "fourfold": ["valine", "alanine", "glycine", "proline", "threonine"],
                   "sixfold": ["leucine", "serine", "arginine"]}


    ### counts of codons within an amino acid
    n_a = {"isoleucine": 0.0, "leucine": 0.0, "valine": 0.0, "phenylalanine": 0.0, "methionine": 0.0,
                    "cysteine": 0.0, "alanine": 0.0, "glycine": 0.0, "proline": 0.0, "threonine": 0.0, "serine": 0.0,
                    "tyrosine": 0.0, "tryptophan": 0.0, "glutamine": 0.0, "asparagine": 0.0, "histidine": 0.0,
                    "glutamic acid": 0.0, "aspartic acid": 0.0, "lysine": 0.0, "arginine": 0.0}

    ### homozygosity
    F_a = {"isoleucine": 0.0, "leucine": 0.0, "valine": 0.0, "phenylalanine": 0.0, "methionine": 0.0,
                    "cysteine": 0.0, "alanine": 0.0, "glycine": 0.0, "proline": 0.0, "threonine": 0.0, "serine": 0.0,
                    "tyrosine": 0.0, "tryptophan": 0.0, "glutamine": 0.0, "asparagine": 0.0, "histidine": 0.0,
                    "glutamic acid": 0.0, "aspartic acid": 0.0, "lysine": 0.0, "arginine": 0.0}


    ### define degeneracy classes
    F_d = {"onefold": 0.0, "twofold": 0.0, "threefold": 0.0, "fourfold": 0.0, "sixfold": 0.0}

    Nc = 0.0

    ### loop through degeneracy classes
    for d in degen_class:

        if verbose:
            print "START OF A DEGENERACY CLASS"
            print "degeneracy class is: " + str(d)


        aminos_in_degen_class = degen_class[d]
        non_zero_amino_counts = 0.0

        for am_ac in aminos_in_degen_class:

            n_a[am_ac] = am_count[am_ac] #cod_count[cod]

            if verbose:
                print "amino acid is: " + str(am_ac)
                print "na_[" + str(am_ac) + "] " + str(n_a[am_ac])

            #print "na_[am_am]: %s" % str(n_a[am_ac])

            cod_list = codons_in_amino(am_ac)

            if n_a[am_ac] > 1: ### means F_a is defined

                for cod in cod_list:
                    #print "codon is %s" % cod
                    #print "observed frequency: %s" % str(c_freq[cod])
                    #print "expec_codon_freq: %d" % expec_cod_freq(cod,n_freqs)
                    F_a[am_ac] += c_freq[cod]**2

                F_a[am_ac] *= n_a[am_ac]
                F_a[am_ac] = F_a[am_ac] - 1
                F_a[am_ac] = F_a[am_ac]/(n_a[am_ac] - 1.0)


            if verbose:
                print "F_a[" + str(am_ac) + "]: " + str(F_a[am_ac])


            #print "F_a[am_ac]: %s" % str(F_a[am_ac])
            if F_a[am_ac] != 0.0:
                F_d[d] += F_a[am_ac]
                non_zero_amino_counts += 1.0

            if verbose:
                print "number of different observed amino_acids in degeneracy class " + str(d) + " is: " + str(non_zero_amino_counts)


        if non_zero_amino_counts > 0.0:
            F_d[d] = F_d[d] / non_zero_amino_counts
        else:
            F_d[d] = 0.0



        ### Novembre: "if k-fold redundancy classes are unobserved, F_d[d] is assumed to be 1/k (k is number of amino acids of same degeneracy class)"
        if F_d[d] != 0:
            Nc += len(degen_class[d]) / F_d[d]
        elif F_d["threefold"] == 0.0: ### Novembre: "if F_3["threefold"] is unobserved, it should be estimated by averaging F_2 and F_4.
            Nc += 0.0
        elif F_d[d] == 0.0: ### this will not be executed if first looped through "threefold"

            ### This is ugly code: correct it! float(degen_amino(am_ac)) is onl defined in inner for loop above
            Nc += len(degen_class[d])*float(degen_amino(am_ac))


    ### Novembre: "if F_3["threefold"] is unobserved, it should be estimated by averaging F_2 and F_4.
    if F_d["threefold"] == 0.0 and F_d["twofold"] > 0.0 and F_d["fourfold"] > 0.0 and F_d["sixfold"] > 0.0:

        if verbose:
            print "averaging procedure for threefold degeneracy"

        ### As suggested by Wright:
        #F_three_avg = (F_d["twofold"] + F_d["fourfold"])/2.0
        ### As suggested by Fulsang:
        F_three_avg = 1/3.0*((2/F_d["twofold"] -1)**(-1.0) + (2/(3.0*F_d["fourfold"]) - 1.0/3.0)**(-1.0) + (2/(5.0*F_d["sixfold"]) - 3.0/5.0)**(-1.0) )
        Nc += len(degen_class["threefold"]) / F_three_avg

    ### if nothing works, simply replace by "crude" frequency of usage of codons
    if Nc == 0.0:
        ### replace simply by number of used codons
        #sum([1 for x in c_freq if c_freq[x] > 0]) does not seem to count the stop codons as well
        Nc = sum([1.0 for x in c_freq if c_freq[x] > 0])

    return Nc

### test
#test_seq = whole_data[1]["mrna_seq"].replace("T", "U")
#Nc(test_seq)
#ribo_seq_test = retrieve_ribo_seq(whole_data)
#ribo_seq_test = ribo_seq_test.replace("T", "U")
#Nc(ribo_seq_test)

### computes amino acid frequencies from a codon count dictionary
### This corresponds to the mathematical object F_a in the calculation of B
def amino_freq(cod_count):

    amino_acid_frequencies = {}

    amino_list = ["isoleucine", "leucine", "valine","phenylalanine", "methionine",
                    "cysteine", "alanine", "glycine", "proline", "threonine", "serine",
                    "tyrosine", "tryptophan", "glutamine", "asparagine", "histidine",
                    "glutamic acid", "aspartic acid", "lysine", "arginine"]

    cod_nmb = sum(cod_count.values())

    for amino_key in amino_list:

        a_freq = 0.0

        ### sum of al codon counts decoding for the same amino acid
        for cod_key in codons_in_amino(amino_key):

            ### this is a count now, but will
            ### be converted to a frequency by
            ### normalizing
            a_freq += cod_count[cod_key]


        ### to obtain a frequency, the number is normalized
        a_freq = a_freq/float(cod_nmb)

        #### fill in the frequency
        amino_acid_frequencies[amino_key] = a_freq

    return amino_acid_frequencies

### testing
# cod_example = codon_count(whole_data[1]["mrna_seq"].replace("T", "U"))
# print amino_freq(cod_example)


### computes amino acid counts from a codon count dictionary
### This corresponds to the mathematical object F_a in the calculation of B
def amino_count(cod_count):

    amino_acid_count = {}

    amino_list = ["isoleucine", "leucine", "valine","phenylalanine", "methionine",
                    "cysteine", "alanine", "glycine", "proline", "threonine", "serine",
                    "tyrosine", "tryptophan", "glutamine", "asparagine", "histidine",
                    "glutamic acid", "aspartic acid", "lysine", "arginine"]

    for amino_key in amino_list:

        a_count = 0.0

        ### sum of al codon counts decoding for the same amino acid
        for cod_key in codons_in_amino(amino_key):

            ### this is a count now, but will
            ### be converted to a frequency by
            ### normalizing
            a_count += cod_count[cod_key]


        #### fill in the frequency
        amino_acid_count[amino_key] = a_count

    return amino_acid_count

### testing
# cod_example = codon_count(whole_data[1]["mrna_seq"].replace("T", "U"))
# print amino_count(cod_example)



### This function converts a dictionary of
### of counts of codons within some sequence into a
### dictionary of frequencies of those counts.
### These frequencies are given relative to the sum of all counts within codon
### group decoding for the same amino acid
### This corresponds to "f_ac" in in Roth, Anisimova, Cannarozzi 2012 ("Measuring Codon Usage Bias")
def cod_freq(cod_count):


    amino_list = ["isoleucine", "leucine", "valine", "phenylalanine", "methionine",
                  "cysteine", "alanine", "glycine", "proline", "threonine", "serine",
                  "tyrosine", "tryptophan", "glutamine", "asparagine", "histidine",
                  "glutamic acid", "aspartic acid", "lysine", "arginine"]

    codon_frequency = {}

    ###
    for amino_key in amino_list:

        a_count = 0.0

        ### sum of al codon counts decoding for the same amino acid
        for cod_key in codons_in_amino(amino_key):

            ### this is a count now, but will
            ### be converted to a frequency by
            ### normalizing
            a_count += cod_count[cod_key]

        for cod_key in codons_in_amino(amino_key):

            if a_count > 0:

                codon_frequency[cod_key] = cod_count[cod_key]/float(a_count)
            else:
                ### what to do when there are no codons decoding for the same amino_acid
                ### does that make sense? Since it is used to determine distances between counts, I assume it is.
                codon_frequency[cod_key] = 0.0

    return codon_frequency


### test
# cod_count_test = codon_count(whole_data[1]["mrna_seq"].replace("T", "U"))
# cod_freq(cod_count_test)



### calculates the frequency of each codon in an mRNA sequence
def codon_freq_in_amino(sequence, finish_with_stop = False):

    ### dictionary with all codons, and their respective counts
    cod_count = {"UAG": 0.0, "UAA": 0.0, "UGA": 0.0, "AUU": 0.0, "AUC": 0.0, "AUA": 0.0, "CUU": 0.0, "CUC": 0.0, "CUA": 0.0, "CUG": 0.0,
                 "UUA": 0.0, "UUG": 0.0,
                 "GUU": 0.0, "GUC": 0.0, "GUA": 0.0, "GUG": 0.0, "UUU": 0.0, "UUC": 0.0, "AUG": 0.0, "UGU": 0.0, "UGC": 0.0, "GCU": 0.0,
                 "GCC": 0.0, "GCA": 0.0,
                 "GCG": 0.0, "GGU": 0.0, "GGC": 0.0, "GGA": 0.0, "GGG": 0.0, "CCU": 0.0, "CCC": 0.0, "CCA": 0.0, "CCG": 0.0, "ACU": 0.0,
                 "ACC": 0.0, "ACA": 0.0,
                 "ACG": 0.0, "UCU": 0.0, "UCC": 0.0, "UCA": 0.0, "UCG": 0.0, "AGU": 0.0, "AGC": 0.0, "UAU": 0.0, "UAC": 0.0, "UGG": 0.0,
                 "CAA": 0.0, "CAG": 0.0,
                 "AAU": 0.0, "AAC": 0.0, "CAU": 0.0, "CAC": 0.0, "GAA": 0.0, "GAG": 0.0, "GAU": 0.0, "GAC": 0.0, "AAA": 0.0, "AAG": 0.0,
                 "CGU": 0.0, "CGC": 0.0,
                 "CGA": 0.0, "CGG": 0.0, "AGA": 0.0, "AGG": 0.0}

    ### set up a similar dictionary
    cod_freq = cod_count

    ### go through sequence and add count
    seq_length = len(sequence)
    ### seq_length should be divisible by three
    L = seq_length / 3

    ### what do we do if it is not? In fact, often times it is not divisible by three.
    tot_count = 0.0


    for k in range(0, L, 1):
        cod = sequence[3 * k:3 * k + 3]
        if genetic_code(cod) == "stop" and finish_with_stop:
            print "stop codon found at site %d" % float(3 * k)
            break
        cod_count[cod] += 1
        tot_count += 1.0


    ### define a library of the codons decoding for each amino acid
    amino_codon_list = ["stop", "isoleucine", "leucine", "valine", "phenylalanine",
                           "methionine", "cysteine", "alanine", "glycine",
                           "proline", "threonine", "serine", "tyrosine",
                           "tryptophan", "glutamine", "asparagine",
                           "histidine", "glutamic acid", "aspartic acid",
                           "lysine", "arginine"]

    cod_fam_count = {"stop": 0.0, "isoleucine": 0.0,
                           "leucine": 0.0, "valine": 0.0, "phenylalanine": 0.0,
                           "methionine": 0.0, "cysteine": 0.0, "alanine": 0.0,
                           "glycine": 0.0, "proline": 0.0, "threonine": 0.0,
                           "serine": 0.0, "tyrosine": 0.0, "tryptophan": 0.0,
                           "glutamine": 0.0, "asparagine": 0.0, "histidine": 0.0,
                           "glutamic acid": 0.0, "aspartic acid": 0.0, "lysine": 0.0,
                           "arginine": 0.0}


    ### create the sum of the codons within a codon family of an amino acid
    for am in amino_codon_list:
        for cod_am in codons_in_amino(am):
            cod_fam_count[am] += cod_count[cod_am]



    ### normalize with the total number of codons within a family
    for cod in cod_count:
        if cod_fam_count[genetic_code(cod)] > 0.0:
            cod_freq[cod] = cod_count[cod]/cod_fam_count[genetic_code(cod)]
        else:
            cod_freq[cod] = 0.0


    return cod_freq

### test
# test_seq = slip_data[1]["sequence"]
# test_seq = whole_data[1]["mrna_seq"].replace("T", "U")
# codon_freq_in_amino(test_seq)



### find the index in whole_data that corresponds to a particular accession
### number in slip_data.
def retrieve_wd_index(whole_data, acc_name):

    whole_len = len(whole_data)

    acc_index = 0

    for ind in range(0, whole_len):

        if whole_data[ind]["accession"] == acc_name :
            acc_index = ind
            break

    return acc_index


### find the index in whole_data that corresponds to a particular gene name
### in the data.
def retrieve_wd_geneind(whole_data, gene_name):

    whole_len = len(whole_data)

    gene_index = 0

    for ind in range(0, whole_len):

        #print whole_data[ind]["accession"][-5:]

        ### must exclude the data from the W303 strain. These are characterized by an ending _W303 in the accession feature of the genome dataset
        if whole_data[ind]["genename"] == gene_name and whole_data[ind]["accession"][-5:] != "_W303":
            gene_index = ind
            break

    return gene_index



### fuse the original with a frameshifted sequence
def fuse_seq(org_seq, slip_start, slip_seq):

    fused_seq = ''


    ### the slippery site is at slip_start to slip_start + 7
    ### the last nucleotide in the 7-nucleotide slippery site
    ### is the first nucleotide of the frame-shifted reading frame
    #pre_PRF = org_seq[0:(slip_start + 7)] ### should include the last nucleotide of the ORF (at slip_start + 6)
    pre_PRF = org_seq[0:(slip_start + 6)] ### the above code should be correct, but somehow the start of the slip site seems to be off by one
    post_PRF = slip_seq

    if post_PRF is not None:
        fused_seq = pre_PRF + post_PRF
    else:
        fused_seq = pre_PRF

    return fused_seq



### obtain the original sequence until a slippery site (until some slip_start + 6)
def seq_until_slipsite(org_seq, slip_start):

    ### the slippery site is at slip_start to slip_start + 7
    ### the last nucleotide in the 7-nucleotide slippery site
    ### is the first nucleotide of the frame-shifted reading frame
    #pre_PRF = org_seq[0:(slip_start + 7)] ### should include the last nucleotide of the ORF (at slip_start + 6)
    pre_PRF = org_seq[0:(slip_start + 6)] ### the above code should be correct, but somehow the start of the slip site seems to be off by one

    return pre_PRF


### obtain the original sequence after a slippery site (after some slip_start + 6) until the stop codon
def ORFseq_after_slipsite(org_seq, slip_start):

    ### the slippery site is at slip_start to slip_start + 7
    ### the last nucleotide in the 7-nucleotide slippery site
    ### is the first nucleotide of the frame-shifted reading frame
    #pre_PRF = org_seq[0:(slip_start + 7)] ### should include the last nucleotide of the ORF (at slip_start + 6)
    #pre_PRF = org_seq[0:(slip_start + 6)] ### the above code should be correct, but somehow the start of the slip site seems to be off by one
    post_PRF = org_seq[(slip_start + 6):len(org_seq)] ### thus, the part of the ORF that remains must be like htis

    return post_PRF


### fuse the original with a frameshifted sequence
def effective_slip_seq(org_seq, slip_start, slip_seq):

    ### the last nucleotide of the original sequence org_seq should be added to the slippery sequency

    fused_seq = ''


    ### the slippery site is at slip_start to slip_start + 7
    ### the last nucleotide in the 7-nucleotide slippery site
    ### is the first nucleotide of the frame-shifted reading frame
    #slippery_site = org_seq[(slip_start-1):(slip_start + 6)]
    #pre_PRF = org_seq[0:(slip_start + 7)] ### should include the last nucleotide of the ORF (at slip_start + 6)
    pre_PRF = org_seq[(slip_start + 6) - 1] ### the above code should be correct, but somehow the start of the slip site seems to be off by one
    post_PRF = slip_seq

    if post_PRF is not None:
        fused_seq = pre_PRF + post_PRF
    else:
        fused_seq = ''

    return fused_seq


### compute the frameshifted part of the
### sequence that is translated after programmed
### ribosomal frameshifting
def get_slip_sequence(org_seq, slip_start):

    prf_seq = ''

    ### this is the one nucleotide that remains from the ORF
    ### the first nucleotide of the frameshifted sequence
    pre_PRF = org_seq[(slip_start + 6) - 1]
    slippery_site = org_seq[(slip_start - 1):(slip_start + 6)]
    post_PRF = org_seq[(slip_start + 6):len(org_seq)]

    prf_seq = pre_PRF + post_PRF

    ### cutoff at a stop codon
    ### go through sequence and add count
    seq_length = len(prf_seq)

    ### seq_length should be divisible by three
    L = seq_length / 3 ### not taking division by 3.0 on purpose

    ### the last codon index, in number of codons
    end_cod_nmb = L

    for k in range(0, L, 1):

        cod = prf_seq[3 * k: 3 * k + 3]
        if genetic_code(cod) == "stop":
            print "stop codon found at site %d" % float(3 * k)
            end_cod_nmb = k
            break

    ### exclude the last codon nucleotides that do not fall into a codon
    prf_seq = prf_seq[0: 3*end_cod_nmb]

    return prf_seq



### now loop through all accession names, and compute the CAI for that gene in
### in original ORF and in the PRF-1 reading frame
def compute_indices(slip_data, whole_data, n_freqs):

    ### define return object
    ind_vals = {}

    ### number of entries from S288C
    slip_len = len(slip_data)


    ### retrieve ribosome gene sequences
    ribo_seq = retrieve_ribo_seq(whole_data)

    ### replace T by U
    ribo_seq = ribo_seq.replace("T", "U")

    ribo_seq_relad = rel_adaptiveness(ribo_seq)

    ribo_amino_freq = amino_freq(codon_count(ribo_seq))
    ribo_cod_freq = cod_freq(codon_count(ribo_seq))

    ### loop through the accession names of the genes with a slippery site
    for acc in range(0, slip_len):

        ### retrieve the index where the accession number in slip_data is
        ### stored in the whole_data
        acc_name = slip_data[acc]["accession"]

        print "looping through accession name: " + acc_name + " at number " + str(acc)

        ###############################################
        # open reading frame sequence computations
        ###############################################

        print "compute indices for ORF"


        whole_dt_ind = retrieve_wd_index(whole_data, acc_name)
        wh_seq = whole_data[whole_dt_ind]["mrna_seq"]

        ### this is pre-transcription sequence. Need to transform to RNA sequence
        wh_seq = wh_seq.replace("T", "U")

        ### Compute indices for the non-frameshifted data

        ### CAI


        print "CAI ORF"

        ### compute the CAI on the original sequence
        cai_whole = CAI_f(wh_seq, ribo_seq_relad)


        ### B

        print "B ORF"

        whole_cod_freq = cod_freq(codon_count(wh_seq))

        B_whole = B(whole_cod_freq, ribo_cod_freq, ribo_amino_freq)

        ### Nc prime

        print "Nc\' ORF"

        Ncprime_whole = Ncprime(wh_seq, n_freqs)

        #if Ncprime_whole > 61:
        #    print "erroneous Nc\' calculation att acc %s" % str(acc)
        #    break

        ### Nc

        print "Nc ORF"

        Nc_whole = Nc(wh_seq)

        ###############################################
        # PRF -1 sequence computations (original sequence until slippery site, and adding the prf-sequence until the stop codon)
        ###############################################

        print "compute indices for PRF -1"

        ### auxiliary parms
        slip_start = slip_data[acc]["start"]
        #prf_seq = slip_data[acc]["sequence"] ### wrong
        prf_seq = get_slip_sequence(wh_seq,slip_start)
        slip_site = slip_data[acc]["slipsite"] ### not used at the moment

        slip_seq = ''

        print "testing whether prf_seq makes sense"
        print "is prf_seq equal to None? " + str(prf_seq is None)
        if prf_seq is not None:

            print "fusing wh_seq and prf_seq"

            print "prf_seq is:" + str(prf_seq)

            ### construct computed slip sequence
            slip_seq = fuse_seq(wh_seq, slip_start, prf_seq)
        else:
            print "prf_seq is empty"
            slip_seq = wh_seq
            prf_seq = ''

        ### Compute indices for the frameshifted sequence


        ### CAI

        print "CAI PRF"

        ### compute the CAI on the fused sequence
        cai_slip = CAI_f(slip_seq, ribo_seq_relad)


        ### Computed corrected CAI value

        m_pi = 0.0

        p = 0.01
        cai_est_1percent = (1.0/(1.0 + (float(slip_start + len(prf_seq) + m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole

        p = 0.05
        cai_est_5percent = (1.0/(1.0 + (float(slip_start + len(prf_seq)+ m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole

        p = 0.1
        cai_est_10percent = (1.0/(1.0 + (float(slip_start + len(prf_seq)+ m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole

        p = 0.3
        cai_est_30percent = (1.0/(1.0 + (float(slip_start + len(prf_seq)+ m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole

        p = 0.5
        cai_est_50percent = (1.0/(1.0 + (float(slip_start + len(prf_seq)+ m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole

        p = 0.9
        cai_est_90percent = (1.0/(1.0 + (float(slip_start + len(prf_seq)+ m_pi)/float(len(wh_seq)) )*(p/(1-p)) ))*cai_whole


        ### B

        ###

        print "B PRF"

        slip_cod_freq = cod_freq(codon_count(slip_seq))

        ### compute B for slip _seq
        B_slip = B(slip_cod_freq, ribo_cod_freq, ribo_amino_freq)


        ### Nc prime

        print "Nc\' PRF"

        Ncprime_slip = Ncprime(slip_seq, n_freqs, False)


        ### Nc

        print "Nc PRF"

        Nc_slip = Nc(slip_seq, False)

        ###############################################
        # BIAS_0 (original sequence until slippery site)
        ###############################################

        print "compute indices for sequence until slippery sites"

        ### auxiliary parms
        #slip_start = slip_data[acc]["start"]
        seq_0 = seq_until_slipsite(wh_seq, slip_start)

        ### Compute indices for the frameshifted sequence


        ### CAI

        print "CAI until slippery site"

        ### compute the CAI on the fused sequence
        cai_0 = CAI_f(seq_0, ribo_seq_relad)

        print "CAI until slippery site withouth initial ramp of 10 codons"


        ### rampless sequence until slippery site
        if slip_start > 10*3:
            seq_rampless_10 = seq_until_slipsite(wh_seq[10*3::], slip_start)
            cai_0_rampless_10 = CAI_f(seq_rampless_10, ribo_seq_relad)
        else:
            cai_0_rampless_10 = -1.0


        print "CAI until slippery site withouth initial ramp of 45 codons"


        ### rampless sequence until slippery site
        if slip_start > 45*3:
            seq_rampless_45 = seq_until_slipsite(wh_seq[45*3::], slip_start)
            cai_0_rampless_45 = CAI_f(seq_rampless_45, ribo_seq_relad)
        else:
            cai_0_rampless_45 = -1.0



        print "CAI until slippery site withouth initial ramp of 100 codons"


        ### rampless sequence until slippery site
        if slip_start > 100*3:
            seq_rampless_100 = seq_until_slipsite(wh_seq[100*3::], slip_start)
            cai_0_rampless_100 = CAI_f(seq_rampless_100, ribo_seq_relad)
        else:
            cai_0_rampless_100 = -1.0


        ### B

        print "B until slippery site"

        seq0_cod_freq = cod_freq(codon_count(seq_0))

        ### compute B for slip _seq
        B_0 = B(seq0_cod_freq, ribo_cod_freq, ribo_amino_freq)

        ### Nc prime

        print "Nc\' until slippery site"

        Ncprime_0 = Ncprime(seq_0, n_freqs, False)

        ### Nc

        print "Nc until slippery site"

        Nc_0 = Nc(seq_0, False)


        ###############################################
        # BIAS_1 (original sequence after slippery site until natural end)
        ###############################################


        print "compute indices for ORF sequence after slippery sites"

        ### auxiliary parms
        #slip_start = slip_data[acc]["start"]
        seq_1 = ORFseq_after_slipsite(wh_seq, slip_start)

        ### Compute indices for the frameshifted sequence


        ### CAI

        print "CAI after slippery site"

        ### compute the CAI on the fused sequence
        cai_1 = CAI_f(seq_1, ribo_seq_relad)

        ### B

        print "B after slippery site"

        seq1_cod_freq = cod_freq(codon_count(seq_1))

        ### compute B for slip _seq
        B_1 = B(seq1_cod_freq, ribo_cod_freq, ribo_amino_freq)

        ### Nc prime

        print "Nc\' after slippery site"

        Ncprime_1 = Ncprime(seq_1, n_freqs, False)

        ### Nc

        print "Nc after slippery site"

        Nc_1 = Nc(seq_1, False)

        ###############################################
        # BIAS_2 (translated PRF sequence after slippery site until premature end)
        ###############################################


        print "compute indices for ORF sequence after slippery sites"

        ### auxiliary parms
        # slip_start = slip_data[acc]["start"]
        seq_2 = effective_slip_seq(wh_seq, slip_start, slip_seq)

        ### Compute indices for the frameshifted sequence


        ### CAI

        print "CAI after slippery site"

        ### compute the CAI on the fused sequence
        cai_2 = CAI_f(seq_2, ribo_seq_relad)

        ### B

        print "B after slippery site"

        seq2_cod_freq = cod_freq(codon_count(seq_2))

        ### compute B for slip _seq
        B_2 = B(seq2_cod_freq, ribo_cod_freq, ribo_amino_freq)

        ### Nc prime

        print "Nc\' after slippery site"

        Ncprime_2 = Ncprime(seq_2, n_freqs, False)

        ### Nc

        print "Nc after slippery site"

        Nc_2 = Nc(seq_2, False)


        ###############################################
        # combine results
        ###############################################

        ### some of the accession names appear more than once,
        ### and will therefore be overwritten

        ### store the results
        ind_vals[acc_name +"_"+ str(slip_start)] = {"CAI_ORF" : cai_whole, "CAI_PRF": cai_slip,
                            "CAI_1perc_p": cai_est_1percent, "CAI_5perc_p": cai_est_5percent,
                            "CAI_10perc_p": cai_est_10percent, "CAI_30perc_p": cai_est_30percent,
                            "CAI_50perc_p": cai_est_50percent, "CAI_90perc_p": cai_est_90percent,
                            "B_ORF": B_whole, "B_PRF": B_slip,
                            "Ncprime_ORF": Ncprime_whole, "Ncprime_PRF": Ncprime_slip, "Nc_ORF": Nc_whole, "Nc_PRF": Nc_slip,
                            "CAI_0": cai_0, "CAI_0_noramp10": cai_0_rampless_10, "CAI_0_noramp45": cai_0_rampless_45, "CAI_0_noramp100": cai_0_rampless_100,
                            "CAI_1": cai_1, "CAI_2": cai_2,  "B_0": B_0, "B_1": B_1, "B_2": B_2,
                            "Nc_0": Nc_0, "Nc_1": Nc_1, "Nc_2": Nc_2, "Ncprime_0": Ncprime_0, "Ncprime_1": Ncprime_1, "Ncprime_2": Ncprime_2,
                            "length_whseq": len(wh_seq), "length_slipseq": len(slip_seq), "length_prfseq": len(prf_seq), "slip_start": slip_start,
                            "genename": whole_data[whole_dt_ind]["genename"], "accession": acc_name}

    return ind_vals




########################################################################################################################
################ COMPUTATIONS ################
########################################################################################################################


### put all the indices into one single object
indices = compute_indices(slip_data, whole_data, nucl_freqs)

### store everyting in a jason file
with open('my_data.json', 'w') as fp:
    json.dump(indices, fp)

fp.close()

### test whether all of the data have been used
len(indices) == len(slip_data)


########################### CLOSE THE CONNECTION TO THE CURSOR #######################################
cursor.connection.commit()


cursor.close()
connection.close()



