### HERE WE DEFINE THE FUNCTIONS THAT PLOT THE FIGURES IN THE SLIPPERY PROJECT MANUSCRIPT.
###
### Victor Garcia, Zurich University of Applied Sciences (ZHAW), October 2017
########################################################################################################################

import json
import matplotlib.pyplot as plt
import pandas as pd
import colorbrewer
import numpy as np
from scipy import stats
import seaborn as sns
import ggplot
from statsmodels.regression import linear_model
from statsmodels.api import add_constant

#import plotly.plotly as py

####### load the data generated in analyse_slippery_sites.py
with open('my_data.json', 'r') as fp:
    indices = json.load(fp)

### test whether correct:
indices[indices.keys()[0]]


### auxiliary function to convert numbers into scientific notation string
def format_e(n):
    a = '%E' % n
    mola = round(float(a.split('E')[0].rstrip('0').rstrip('.')), 3)
    return str(mola) + 'E' + a.split('E')[1]



########################################################################################################################



########################################################################################################################
################ FIGURE 1: CONCEPT PLOT FOR MATHEMATICAL NOTATION USED IN THE MANUSCRIPT ################
########################################################################################################################

# This is a concept figure which was built with keynote software from Apple Inc.



#######################################################################################################################
################ FIGURE 2: THE COST OF THE -1 PRF MECHANISM ################
########################################################################################################################

#######################################################################################################################
################ FIGURE 2A: DECLINE IN THE NUMBER OF SLIPPERY SITES WITH PROTEIN EXPRESSION LEVELS ################
########################################################################################################################


#### Ghaemmaghani et. al, 2003, Nature dataset

### read in the table
#ghaemm_table = pd.read_table("ghaemmaghani2003n_SUPPdata.txt")

### test
#print ghaemm_table['ORF'][0]
#print ghaemm_table['Observed Expression'][0]
#print ghaemm_table['Protein Molecules/Cell'][0]
#print ghaemm_table['Standard Deviation'][0]
#print ghaemm_table['CEC'][0]
#print ghaemm_table['Spurious'][0]


#### von der Haar, 2008, BMC Systems Biology curated dataset

### read in the table
vanderHaar_table = pd.read_csv("vanderHaar_curated.csv", sep = ";")

### test
#print vanderHaar_table['ORF'][0]
#print vanderHaar_table['Gene Name'][0]
#print vanderHaar_table['Length (Amino Acids)'][0]
#print vanderHaar_table['Protein Abundance (molecules per cell)'][0]
#print vanderHaar_table['Protein decay rate (min-1)'][0]
#print vanderHaar_table['mRNA Abundance (molecules per cell)'][0]



### load paxdb whole integrated yeast genome dataset, suggested by Adam Hockenberry

paxdb_table = pd.read_table("Scerv_whole_integrated_paxdb.txt")
#paxdb_table_copy = paxdb_table

#print paxdb_table['ORF'][0]
#print paxdb_table['abundance'][0]

### problem is that ORF is given with a prefix. Need to remove prefix.

### proper way to write into a pandas table is this:
#paxdb_table.set_value(paxdb_table.index[0], "ORF", 'test')

### add the ORF name to the paxdb table
for i in range(0,len(paxdb_table)):
    orf_name = paxdb_table['ORF'][i][5:]
    #paxdb_table['ORF'][i] = orf_name
    ### need to use special function to set values
    paxdb_table.set_value(paxdb_table.index[i], "ORF", orf_name)


### function that, given a accession will return the mRNA level from the Holstege data set.
def mrna_lev_of_gene(holstege_datatable, gene_name, keyname_of_gene = 'Gene', verbose = False):

    ### the list object with the result (the mRNA level) of the gene to be returned to the user
    mrna_level = 0.0

    ### number of data entries in the Holstege data set
    row_num = len(holstege_datatable)

    if verbose:
        print "entering for loop for search"

    for i in range(0, row_num):

        if verbose:
            print str(i)
            print str(holstege_datatable[keyname_of_gene][i])


        ### if entry in data table is equal to one asked for by user,
        ### return store mRNA level, and return it.
        if holstege_datatable[keyname_of_gene][i] == gene_name:

            mrna_level = holstege_datatable['ExpressionLevel(Copies/Cell)'][i]
            break

    return mrna_level


### retrieves the key within 'indices' dictionary where a gene with 'gene_name' is stored
def retrieve_gene_key(indices, gene_name, keyname_of_gene = 'genename', keyname_of_accession = 'accession', verbose = False):


    ### search for the key in indices that matches the gene_name passed
    ### by the user
    gene_key = []

    for key in indices:
        if verbose:
            print indices[key][keyname_of_gene]
        if indices[key][keyname_of_gene] == gene_name and indices[key][keyname_of_accession][-5:] != '_W303':
            gene_key.append(key)

    return gene_key



### retrieves the key where a gene is stored in the incdices dictionary
def retrieve_accession_key(indices, accession_name, keyname_of_accession = 'accession', verbose = False):


    ### search for the key in indices that matches the gene_name passed
    ### by the user
    acc_key = []

    for key in indices:

        #print indices[key][acc_name_ind][keyname_of_accession]
        if indices[key][keyname_of_accession] == accession_name:
            acc_key.append(key)

    return acc_key

### test
#retrieve_accession_key(indices, 'YAL008W')
#retrieve_accession_key(indices, 'YPR204W')


### retrieves the key where a gene is stored in the incdices dictionary
def retrieve_accession_key(indices, accession_name, keyname_of_accession = 'accession', verbose = False):


    ### search for the key in indices that matches the gene_name passed
    ### by the user
    acc_key = []

    for key in indices:

        #print indices[key][acc_name_ind][keyname_of_accession]
        if indices[key][keyname_of_accession] == accession_name:
            acc_key.append(key)

    return acc_key


### translates an accession name into a corresponding gene name
def accession_to_gene(indices, accession_name, keyname_of_accession = 'accession', verbose = False):

    ### search for the key in indices that matches the gene_name passed
    ### by the user
    gene_name = ""

    for key in indices:

        #print indices[key][acc_name_ind][keyname_of_accession]
        if indices[key][keyname_of_accession] == accession_name:

            gene_name = indices[key]['genename']

            break

    return gene_name

### test
#conv_gene_name = accession_to_gene(indices, 'YKL029C', keyname_of_accession = 'accession', verbose = False)
#print conv_gene_name

### retrieve and ordering of the slippery site
### This function does the following:
### for a given accession name, acc_name, it
### finds all entries in the data set "indices"
### produced by the script analyse_slippery_sites.py,
### that contain information about acc_name.
### These entries are given back to the user
### in an ordered way in the form of dictionary keys,
### with those entries corresponding
### to early (within the gene) slippery sites first,
### and those corresponding to late slipperay sites at the end
### (that is, in ascending order).
def obtain_ordered_PRFsites(acc_name, indices):

    ind_keys = retrieve_accession_key(indices, acc_name)

    nmb_slip_sites = len(ind_keys)

    slip_positions = []

    for i in range(0,nmb_slip_sites):

        slip_positions.append(indices[ind_keys[i]]["slip_start"])


    rel_order = np.argsort(slip_positions)

    ordered_keys = []

    for i in range(0, nmb_slip_sites):
        ordered_keys.append(ind_keys[rel_order[i]])


    return ordered_keys


### test
#obtain_ordered_PRFsites('YPR204W', indices)


### generates an object with protein expression and codon usage bias information
def create_expression_obj(expr_data, indices, protein_key_name = 'Protein Molecules/Cell',
                          mrna_key_name = '', verbose = True):

    ### determine how many rows there are in the data
    rows_data = len(expr_data)

    ### prepare the output variable: a list of all the protein expression
    ### data of a given gene, as well as the codon usage bias indices for that gene
    expr_bias_data = []

    for i in range(0,rows_data):

        if verbose:
            print "looping trough data set row: " + str(i)

        ### get the accession name of the gene
        ### whose protein expression was measured
        acc_name = expr_data['ORF'][i]

        if verbose:
            print "accession name is: " + acc_name


        ### retrieve the keys within indices
        ### for which there exist measurements
        #ind_keys = retrieve_accession_key(indices, acc_name)
        ind_keys = obtain_ordered_PRFsites(acc_name, indices)

        ### the number of data for tha accession
        nmb_data_acc = len(ind_keys)

        #### extract protein expression levels for this accession
        # prot_expr = float(expr_data['Protein Molecules/Cell'][i])
        prot_expr = expr_data[protein_key_name][i]
        prot_expr = str(prot_expr)


        ### obtain mRNA expression levels (copies/cell), if present
        #gene_name = accession_to_gene(indices, acc_name, keyname_of_accession = 'accession', verbose = False)
        #mrna_level = mrna_lev_of_gene(holstege_datatable, gene_name, verbose = False)
        if len(mrna_key_name) > 0:
            mrna_level = expr_data[mrna_key_name][i]
            mrna_level = str(mrna_level)
        else:
            mrna_level = '-1.0'


        if verbose:
            print "prot_expr is: ###" + prot_expr + "-###"
            #print "is prot_expr = \'-\'? " + str(prot_expr == '-')
            #print "length string prot_expr: " + str(len(prot_expr))
            print "the number of slip sites is: " + str(nmb_data_acc)
            print "mrna_level is: ###" + str(mrna_level) + "-###"



        ### only those genes that contain slippery sites should be evaluated
        #if nmb_data_acc > 0:


        ### loop through all slippery site instances
        for j in range(0,nmb_data_acc):

            if verbose:
                print "retrieved key is " + ind_keys[j]

            ### exclude missing values (that is, all genes where no exression could be measured)
            if prot_expr != '-' and prot_expr != '%' and prot_expr != ' -' and prot_expr != '#':

                if verbose:
                    print "convert prot_expr into float"

                prot_expr = float(prot_expr)
                mrna_level = float(mrna_level)

                expr_bias_data.append({'accession': acc_name,
                                "protein_expression": prot_expr,
                                'mrna_level': mrna_level,
                                "slip_site_number": nmb_data_acc,
                                "jth_slippery_site": str(j)}) #,
                                #"measures" : indices[ind_keys[j]]})

                ### here we add all corresponding information from
                ### the indices object to expr_bias_data
                entries_num = len(indices[ind_keys[j]])
                for k in range(0, entries_num):

                    indices_key_name = indices[ind_keys[j]].keys()[k]
                    expr_bias_data[-1][indices_key_name] = indices[ind_keys[j]].values()[k]


    return expr_bias_data


#### CONSTRUCT DERIVED DATA SET OBJECT FOR BETTER EASE OF PLOTTING

### obtain the data
#ghaemm_data = create_expression_obj(ghaemm_table, indices)
vanderhaar_data = create_expression_obj(vanderHaar_table, indices,
                                        protein_key_name = 'Protein Abundance (molecules per cell)',
                                        mrna_key_name = 'mRNA Abundance (molecules per cell)')
paxdb_data = create_expression_obj(paxdb_table, indices, protein_key_name = 'abundance')


### store everyting in a json file
#with open('ghaemm_data.json', 'w') as fp:
#    json.dump(ghaemm_data, fp)
#fp.close()

#### test
#print ghaemm_data[0]
#print ghaemm_data[len(ghaemm_data)-1]#["genename"]
#ghaemm_data[6211]



### generates an object with protein expression and codon usage bias information
def gen_expression_slipnmb_obj(expr_data, indices,
                               protein_key_name = 'Protein Molecules/Cell',
                               mrna_key_name = '', verbose = True):

    ### determine how many rows there are in the data
    rows_data = len(expr_data)

    ### prepare the output variable: a list of all the protein expression
    ### data of a given gene, as well as the codon usage bias indices for that gene
    expr_bias_data = []

    for i in range(0,rows_data):

        if verbose:
            print "looping trough data set row: " + str(i)

        ### get the accession name of the gene
        ### whose protein expression was measured
        acc_name = expr_data['ORF'][i]

        if verbose:
            print "accession name is: " + acc_name


        ### retrieve the keys within indices
        ### for which there exist measurements
        ind_keys = retrieve_accession_key(indices, acc_name)

        if verbose:
            print "extracting the number of slippery sites for that gene/accession"

        ### the number of data for that accession value
        nmb_data_acc = len(ind_keys)

        if verbose:
            print "extracting the protein expression for that gene/accession"


        #### extract protein expression levels for this accession
        # prot_expr = float(expr_data['Protein Molecules/Cell'][i])
        prot_expr = expr_data[protein_key_name][i]
        prot_expr = str(prot_expr)

        ### obtain mRNA expression levels (copies/cell)
        #gene_name = accession_to_gene(indices, acc_name, keyname_of_accession = 'accession', verbose = False)
        #mrna_level = mrna_lev_of_gene(holstege_datatable, gene_name, verbose = False)
        if len(mrna_key_name) > 0:
            mrna_level = expr_data[mrna_key_name][i]
            mrna_level = str(mrna_level)
        else:
            mrna_level = '-1.0'

        if verbose:
            print "extracting the gene length for that gene/accession"


        #### extract protein expression levels for this accession
        if len(ind_keys) > 0:
            length_gene = indices[ind_keys[0]]['length_whseq']

            if verbose:
                print "key is: " + ind_keys[0]
        else:
            length_gene = 0

        if verbose:
            print "prot_expr is: ###" + prot_expr + "-###"
            #print "is prot_expr = \'-\'? " + str(prot_expr == '-')
            #print "length string prot_expr: " + str(len(prot_expr))
            print "the number of slip sites is: " + str(nmb_data_acc)
            print "the length of the gene is: " + str(length_gene)


        avg_rel_slip_site = 0
        #total_slip_site = 0
        #for j in range(0, nmb_data_acc):
        #    total_slip_site += expr_data[ind_keys[i]]['slip']


        ### exclude missing values
        if prot_expr != '-' and prot_expr != '%' and prot_expr != ' -' and prot_expr != '#':

            if verbose:
                print "convert prot_expr into float"

            prot_expr = float(prot_expr)
            mrna_level = float(mrna_level)

            expr_bias_data.append({'accession': acc_name,
                            "protein_expression": prot_expr,
                            'mrna_level': mrna_level,
                            'nmb_slip_sites': nmb_data_acc,
                            'length_gene': length_gene})
                            #,'slip_site_per_nt': float(nmb_data_acc)/float(length_gene)})



    return expr_bias_data


### retrieve the data
#ghaemm_data_slip = gen_expression_slipnmb_obj(ghaemm_table, indices)
vanderhaar_data_slip = gen_expression_slipnmb_obj(vanderHaar_table, indices,protein_key_name = 'Protein Abundance (molecules per cell)',
                                        mrna_key_name = 'mRNA Abundance (molecules per cell)')
paxdb_data_slip = gen_expression_slipnmb_obj(paxdb_table, indices, protein_key_name = 'abundance')




########################################################################################################################
### PLOTTING PROTEIN EXPRESSION VS NUMBER OF SLIPPERY SITES PER GENE
########################################################################################################################


### choose the data set that should be plotted

#data_set = ghaemm_data
data_set = vanderhaar_data
#data_set = paxdb_data

#data_slippery = ghaemm_data_slip
data_slippery = vanderhaar_data_slip
#data_slippery = paxdb_data_slip

### Fig 3A

### protein expression
x = []
### number of slippery sites per gene per unit length of gene
y_slip = []

### fill lists by looping through data set
for i in range(0,len(data_slippery)):
    nmb_sites = data_slippery[i]["nmb_slip_sites"]
    length_gene = data_slippery[i]["length_gene"]
    if nmb_sites > 0:
        prot_exp = data_slippery[i]["protein_expression"]
        mrna_level = data_slippery[i]["mrna_level"]

        if mrna_level > 0.0 and prot_exp > 0.0:
            x.append(prot_exp/mrna_level)
        else:
            x.append(prot_exp)

        y_slip.append(float(nmb_sites)/float(length_gene))


### compute the average in bins of equal distance

max_x = 7 ### maximum protein expression at around 10^7
min_x = 1 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 10.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_avg = []

### loop through bins
for b in range(1, len(bins)):

    ind_within_bin = [z for z in range(0,len(x)) if x[z] >= 10**bins[b-1] and x[z] < 10**bins[b]]

    all_within_bin = [y_slip[z] for z in ind_within_bin]

    y_avg.append(np.mean(all_within_bin))


mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append(10**(bins[z-1] + (bins[z] - bins[z-1])/2.0) )


######### pyplot ###########################################
####### plot with matplotlib ################
### https://matplotlib.org/users/pyplot_tutorial.html
### http://pbpython.com/effective-matplotlib.html


### style
plt.rcParams.update(plt.rcParamsDefault)
#plt.style.available

#plt.style.use('seaborn-white')
plt.style.use('seaborn-whitegrid')

### begin with figure
fig = plt.figure(figsize = [15,6]) ### width, height in inches

### means there will be 1 times 3 panels (1 row, 3 columns),
### and we are concerning ourselves with the first panel
### the counting is 'clockwise' along the rows.
ax = fig.add_subplot(131)

# if we only want to plot one panel, not three: ax = fig.add_subplot(111)
fig.subplots_adjust(top = 0.85)

### scatterplot
plt.plot(x, y_slip, linestyle='', marker='o', markersize = 3, alpha = 0.3)

### non-normalized axis settings
#ax.set_ylim(-1,30)
#$ax.text(10, 32, 'A', fontsize = 20)

ax.set_ylim(0,0.01)
#ax.set_xlim(10,10**7)
ax.text(10, 0.0105, 'A', fontsize = 20)

### set x axis scale to logarithmic
ax.set_xscale('log')

### set y axis label
ax.set_ylabel('number of slippery sites per nt per gene')

if mrna_level > 0.0:
    ax.set_xlabel('protein expression [molecules/cell]/mRNA level [molecules/cell]')
else:
    ax.set_xlabel('protein expression [molecules/cell]')

### binned averages
plt.plot(mid_bins, y_avg, "ro-", label = 'mean number of slippery sites', linewidth = 2)


### fit a linear regression through the data

### remove all data points where protein expression was 0
y_slip = [y_slip[i] for i in range(0, len(x)) if x[i] > 0.0]
x = [x[i] for i in range(0, len(x)) if x[i] > 0.0]

y_slip = np.array(y_slip)
x = np.array(x)
log10_x = np.log10(x)
m_s,b_s = np.polyfit(log10_x, y_slip, 1)
plt.plot([10**1,10**6], [1*m_s + b_s, 6*m_s + b_s], linewidth = 3, color = "blue")

### add a text with the value of the regression coefficient to the panels
ax.text(10**4, 0.008, "slope:" + str(round(m_s, 5)), fontsize = 10)


### alternative way to compute regression with statsmodels package
X = add_constant(log10_x)
model = linear_model.OLS(y_slip, X)
results = model.fit()

print results.pvalues[1]

ax.text(10**4, 0.007, "p-val: " + format_e(results.pvalues[1]), fontsize = 10)


#######################################################################################################################
################ FIGURE 2B: SLIPPERY SITE RELATIVE WITHIN-GENE DISTRIBUTION ################
########################################################################################################################


### need to extend this to non-ghaemmaghami data!!!
y_slip_norm = []

for i in range(0,len(data_set)):
    y_slip_norm.append(data_set[i]["slip_start"]/float(data_set[i]["length_whseq"]))

## Adding an average to the plot
max_x = 1 ### maximum protein expression at around 10^7
min_x = 0 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 6.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_spos_avg_orf = []
y_snorm_avg_orf = []

y_spos_avg_prf = []
y_snorm_avg_prf = []


########################################################################################################################
############### histogram of distribution of relative slippery site positions
########################################################################################################################

### set paramters for plotting
ax = fig.add_subplot(132)
sns.distplot(a = y_slip_norm, bins = bins, hist=True, kde = False, rug= False )
ax.set_xlim(0,1)
ax.set_ylim(0, 3200)


plt.xlabel('relative slippery site position')
plt.ylabel('frequency')


### run statistical analysis to determine whether different from zero

### the mean
rel_slip_pos_mean = np.mean(y_slip_norm)
### standard error
rel_slip_pos_std = np.std(y_slip_norm)
### error around the mean
err_slip_pos_mean = 1.96*rel_slip_pos_std/np.sqrt(len(y_slip_norm))

ax.text(0.25, 250, "mean: " + str(round(rel_slip_pos_mean, 4)) + " +/-" + str(round(err_slip_pos_mean, 4)), fontsize = 10)
#ax.text(0.25, 800, "error around mean (+/-): " + str(round(err_slip_pos_mean, 5)), fontsize = 10)



ax.text(-0.1, 3200*1.05, 'B', fontsize = 20)


#######################################################################################################################
################ FIGURE 2C: SLIPPERY SITE RELATIVE WITHIN-GENE DISTRIBUTION VS PROTEIN EXPRESSION ################
########################################################################################################################


########################################################################################################################
###  REL SLIPPERY SITE POSITION VS EXPRESSION
########################################################################################################################

y_slip_norm = []
x_expression = []

for i in range(0,len(data_set)):
    y_slip_norm.append(data_set[i]["slip_start"]/float(data_set[i]["length_whseq"]))
    x_expression.append(data_set[i]["protein_expression"])

#### compute averages as well
max_x = 7 ### maximum protein expression at around 10^7
min_x = 1 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 20.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_slip_norm_avg = []
y_slip_norm_avg_stderr = []

for b in range(1, len(bins)):

    ind_within_bin = [z for z in range(0,len(x_expression)) if x_expression[z] >= 10**bins[b-1] and x_expression[z] < 10**bins[b]]
    all_within_bin = [y_slip_norm[z] for z in ind_within_bin]
    y_slip_norm_avg.append(np.mean(all_within_bin))
    y_slip_norm_avg_stderr.append(1.96*np.std(all_within_bin)/np.sqrt(len(all_within_bin)))


mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append(10**(bins[z-1] + (bins[z] - bins[z-1])/2.0) )



### plot
ax = fig.add_subplot(133)

#plt.scatter(x_expression, y_slip_norm)
plt.plot(x_expression, y_slip_norm, linestyle='', marker='o', markersize = 2, alpha = 0.3 )

#plt.plot(mid_bins, y_slip_norm_avg, 'ro-', linewidth = 2)
plt.errorbar(mid_bins, y_slip_norm_avg, y_slip_norm_avg_stderr, color = 'red', label = 'mean length', linewidth = 2)
plt.plot([10**min_x,10**max_x], [0.5,0.5], 'k', linewidth = 2)

### visualize entries that are non-relevant for analysis due to low power
plt.errorbar(mid_bins[-6::1], y_slip_norm_avg[-6::1], y_slip_norm_avg_stderr[-6::1], color = 'violet', label = 'mean length', linewidth = 2)


#### statistical analysis


### fit a linear regression through the data
### convert to arrays

y_slip_norm = [y_slip_norm[i] for i in range(0, len(x_expression)) if x_expression[i] > 0.0]
x_expression = [x_expression[i] for i in range(0, len(x_expression)) if x_expression[i] > 0.0]


y_slip_norm = np.array(y_slip_norm)
x_expression = np.array(x_expression)
log10_x_expr = np.log10(x_expression)
m_s,b_s = np.polyfit(log10_x_expr, y_slip_norm, 1)

### add the regression line
#plt.plot(x_expression, log10_x_expr*m_s + b_s, linewidth = 1, color = "blue")
plt.plot([10**min_x,10**max_x], [min_x*m_s + b_s, max_x*m_s + b_s], linewidth = 3, color = "blue")

### compute the p-value for the null that the regression coefficient differs from 0

### add a text with the value of the regression coefficient to the panels
ax.text(10**4, 0.2, "slope:" + str(round(m_s, 5)), fontsize = 10)

### alternative way to compute regression with statsmodels package
X = add_constant(log10_x_expr)
model = linear_model.OLS(y_slip_norm, X)
results = model.fit()


### results.params
#print str(results.pvalues[1])

ax.text(10**4, 0.1, "p-val:" + format_e(results.pvalues[1]), fontsize = 10)

#### end of statistical analysis



plt.ylim(0,1)
plt.xlim(10**min_x,10**max_x)

#ax = plt.gca()
ax.set_xscale('log')

### add text
ax.text(10, 1.05, 'C', fontsize = 20)

### axis labels
plt.xlabel('protein expression [molecules/cell]')
plt.ylabel('relative position of slippery site within gene')

### save figure
plt.savefig('figs/Fig2.pdf')
plt.clf()
plt.close()


########################################################################################################################
################ FIGURE 3: CONSEQUENCES OF -1 PRF ON TRANSLATION EFFICIENCY ################
########################################################################################################################

# Concept figure which was built with keynote software from Apple Inc.



#######################################################################################################################
################ FIGURE 4: CORRECTING THE CODON ADAPTATION INDEX FOR -1 PRF MECHANISM ################
########################################################################################################################


### relationship between p and protein expression
def p_from_protexp(prot_expr, base = 0.1): #, -0.046):

    ### set -1 PRF efficiency to zero at start
    #p = 0.0
    prot_expr = np.array(prot_expr)

    ### p decreases with the inverse of the logarithm of the protein expression
    #p = base/(np.log10(prot_expr))

    ### no modification, p is independent from protein expression
    p = base

    return p


##### start figure plotting. Determine figure styles ###########
plt.rcParams.update(plt.rcParamsDefault)

### choose style
plt.style.use('seaborn-whitegrid')


########################################################################################################################
#### PLOTTING OF EXPRESSION VS. SYNONYMOUS CODON USAGE BIAS INDICES CORRECTED BY TRANSLATION EFFICIENCY CONSIDERATIONS
########################################################################################################################


fig = plt.figure(figsize = [12,5]) ### width, height in inches

### add first subplot
ax = fig.add_subplot(121)

####### expression vs bias indices

dev_meas = 0.75

y = []
x_orf = []
x_prf = []


### set
p = 0.3
m_pi = 0 #10000

for i in range(0,len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_orf.append(data_set[i]["CAI_ORF"])

    ### compute from first principles
    ### ratio of slippery site + cost to gene length
    len_ratio = float(data_set[i]["length_slipseq"] + m_pi) / float(data_set[i]["length_whseq"])
    eff_p = p_from_protexp(y[-1], base = p)
    m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
    x_prf.append(m_fac*x_orf[-1])


### remove missing protein expression data (order of these commands important
x_orf = [x_orf[i] for i in range(0, len(y)) if y[i] > 0.0]
x_prf = [x_prf[i] for i in range(0, len(y)) if y[i] > 0.0]
y = [y[i] for i in range(0, len(y)) if y[i] > 0.0]


### build scatterplot
m_alpha = 0.3
#plt.scatter(x_prf, y, label = 'PRF', alpha = m_alpha) #, edgecolors= 'k')
#axes = plt.scatter(x_orf,y, label = 'ORF', alpha = m_alpha) #, edgecolors= 'k')#, color = "red", edgecolors= 'k')
plt.scatter(x_prf, y, label = r'$I_c(x)$', alpha = m_alpha, edgecolors= 'k')
plt.scatter(x_orf,y, label = r'$I(x)$', alpha = m_alpha, edgecolors= 'k')#, color = "red", edgecolors= 'k')
ax = plt.gca()
ax.set_yscale('log')

plt.xlim(0,1)
#plt.legend('lower right')
plt.legend()
plt.xlabel('Codon Adaptation Index')
plt.ylabel('protein expression [molecules/cell]')

### add the correlations coefficient as text
log10_y = np.log10(y)

m_corr_prf = np.corrcoef(x_prf,log10_y)[0][1]
m_corr_prf = round(m_corr_prf, 2)
plt.text(dev_meas, ax.get_ylim()[0]*50, r'$I_c(x)$ corr = ' + str(m_corr_prf))#, color = 'blue')


m_corr_prf = stats.spearmanr(x_prf,log10_y)[0]
m_corr_prf = round(m_corr_prf, 2)
plt.text(0.6, ax.get_ylim()[0]*5, r'$I_c(x)$ spearman corr = ' + str(m_corr_prf))#, color = 'blue')


m_corr_orf = np.corrcoef(x_orf,log10_y)[0][1]
m_corr_orf = round(m_corr_orf, 2)
plt.text(dev_meas, ax.get_ylim()[0]*20, r'$I(x)$ corr = ' + str(m_corr_orf))#, color = 'red')


m_corr_prf = stats.spearmanr(x_orf,log10_y)[0]
m_corr_prf = round(m_corr_prf, 2)
plt.text(0.6, ax.get_ylim()[0]*2, r'$I(x)$ spearman corr = ' + str(m_corr_prf))#, color = 'blue')


### fit a linear regression through the data
### prf
x_prf = np.array(x_prf)
m_prf,b_prf = np.polyfit(x_prf, log10_y, 1)
plt.plot(x_prf, 10**(x_prf*m_prf + b_prf), linewidth = 3, color = "black")
plt.plot(x_prf, 10**(x_prf*m_prf + b_prf), linewidth = 2)#, color = 'blue' )

### orf
x_orf = np.array(x_orf)
m_orf,b_orf = np.polyfit(x_orf, log10_y, 1)
plt.plot(x_orf, 10**(x_orf*m_orf + b_orf), linewidth = 3, color = "black")
plt.plot(x_orf, 10**(x_orf*m_orf + b_orf), linewidth = 2) #, color = 'red' )

ax.text(-0.1, 10**6.5, 'A', fontsize = 20)



########################################################################################################################
#### PLOTTING OF EXPRESSION VS. SYNONYMOUS CODON USAGE BIAS INDICES
########################################################################################################################

### add a subplot
ax = fig.add_subplot(122)

####### expression vs bias indices


### build average -1 PRF efficiency
dev_meas = 0.75

y = []
x_orf = []
x_prf = []


for i in range(0,len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_orf.append(data_set[i]["CAI_ORF"])
    x_prf.append(data_set[i]["CAI_PRF"])


### remove missing protein expression data (order of these commands important
x_orf = [x_orf[i] for i in range(0, len(y)) if y[i] > 0.0]
x_prf = [x_prf[i] for i in range(0, len(y)) if y[i] > 0.0]
y = [y[i] for i in range(0, len(y)) if y[i] > 0.0]


### convert to numpy array
y = np.array(y)
x_orf = np.array(x_orf)
x_prf = np.array(x_prf)


### build scatterplot
m_alpha = 0.3

### compute an array of effective -1 PRF efficiencies
eff_p = p_from_protexp(y, base = p)

plt.scatter((1-eff_p)*x_orf + eff_p*x_prf, y, label = r'$I_a(x)$', alpha = m_alpha, edgecolors= 'k')
plt.scatter(x_orf,y, label = r'$I(x)$', alpha = m_alpha, edgecolors= 'k')#, color = "red", edgecolors= 'k')
ax = plt.gca()
ax.set_yscale('log')

plt.xlim(0,1)
#plt.legend('lower right')
plt.legend()
plt.xlabel('Codon Adaptation Index')
plt.ylabel('protein expression [molecules/cell]')
#plt.title('Codon Adaptation Index')

### add the correlations coefficient as text
log10_y = np.log10(y)

m_corr_prf = np.corrcoef(x_prf,log10_y)[0][1]
m_corr_prf = round(m_corr_prf, 2)
plt.text(dev_meas, ax.get_ylim()[0]*50, r'$I_a(x)$ corr = ' + str(m_corr_prf))#, color = 'blue')


m_corr_prf = stats.spearmanr(x_prf,log10_y)[0]
m_corr_prf = round(m_corr_prf, 2)
plt.text(0.6, ax.get_ylim()[0]*5, r'$I_a(x)$ spearman corr = ' + str(m_corr_prf))#, color = 'blue')


m_corr_orf = np.corrcoef(x_orf,log10_y)[0][1]
m_corr_orf = round(m_corr_orf, 2)
plt.text(dev_meas, ax.get_ylim()[0]*20, r'$I(x)$ corr = ' + str(m_corr_orf))#, color = 'red')


m_corr_prf = stats.spearmanr(x_orf,log10_y)[0]
m_corr_prf = round(m_corr_prf, 2)
plt.text(0.6, ax.get_ylim()[0]*2, r'$I(x)$ spearman corr = ' + str(m_corr_prf))#, color = 'blue')


### fit a linear regression through the data
### prf
x_prf = np.array(x_prf)
m_prf,b_prf = np.polyfit((1-p)*x_orf + p*x_prf, log10_y, 1)
plt.plot(x_prf, 10**(x_prf*m_prf + b_prf), linewidth = 3, color = "black")
plt.plot(x_prf, 10**(x_prf*m_prf + b_prf), linewidth = 2)#, color = 'blue' )

### orf
x_orf = np.array(x_orf)
m_orf,b_orf = np.polyfit(x_orf, log10_y, 1)
plt.plot(x_orf, 10**(x_orf*m_orf + b_orf), linewidth = 3, color = "black")
plt.plot(x_orf, 10**(x_orf*m_orf + b_orf), linewidth = 2) #, color = 'red' )

ax.text(-0.1, 10**6.5, 'B', fontsize = 20)

### save figure
plt.savefig("figs/Fig4.pdf")


plt.clf()
plt.close()




#######################################################################################################################
################ FIGURE 5: CORRELATION OF THE ESTIMATOR VALUES TO LOG10(PROT_EXPRESSION) VS -1 PRF EFFICIENCY  ################
########################################################################################################################

def p_from_protexp(prot_expr, base = 0.1): #, -0.046):

    ### set -1 PRF efficiency to zero at start
    #p = 0.0
    #prot_expr = np.array(prot_expr)

    ### p decreases with the inverse of the logarithm of the protein expression
    #p = base/(np.log10(prot_expr))

    ### no modification, p is independent from protein expression
    p = base

    return p



#### start plotting ##############

plt.rcParams.update(plt.rcParamsDefault)
#plt.style.available

plt.style.use('bmh')

### make a two-panel plot out of it
fig = plt.figure(figsize = [12,5]) ### width, height in inches
ax = fig.add_subplot(121)


####### expression vs bias indices
dev_meas = 0.75

### declare variables
y = []
x_prf = []

### run a first round, not to repeat below
for i in range(0, len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_prf.append(data_set[i]["CAI_PRF"])

### convert to np.array
y = np.array(y)
x_prf = np.array(x_prf)

### remove non-zero entries
non_zero_entries = [i for i in range(0, len(y)) if y[i] > 0.0]
x_prf = x_prf[non_zero_entries]
y = y[non_zero_entries]

### compute log_10 of protein expression
log10_y = np.log10(y)

### choose all possible values of parameter pi
m_pi = [0, 10, 100, 1000, 10000]

### loop through possible costs of NMD pathway (measured in nucleotide lengths) m_pi
for e in m_pi:

    ### declare variables
    x_orf = []
    x_1perc_eff = []
    x_5perc_eff = []
    x_10perc_eff = []
    x_30perc_eff = []
    x_50perc_eff = []
    x_90perc_eff = []



    for i in range(0,len(data_set)):
        orf_cai = data_set[i]["CAI_ORF"]
        x_orf.append(orf_cai)

        ### ratio of slippery site + cost to gene length
        len_ratio = float(data_set[i]["length_slipseq"] + e)/float(data_set[i]["length_whseq"])

        ### expression level
        prot_expr = data_set[i]["protein_expression"]

        ### fix the -1 PRF efficiency
        p = 0.01
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0/(1.0 + len_ratio*(eff_p/(1-eff_p)))
        x_1perc_eff.append(m_fac*orf_cai)

        p = 0.05
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_5perc_eff.append(m_fac*orf_cai)

        p = 0.1
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_10perc_eff.append(m_fac*orf_cai)

        p = 0.3
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_30perc_eff.append(m_fac*orf_cai)

        p = 0.5
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_50perc_eff.append(m_fac*orf_cai)

        p = 0.9
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_90perc_eff.append(m_fac*orf_cai)

    ### convert to array
    x_orf = np.array(x_orf)
    x_1perc_eff = np.array(x_1perc_eff)
    x_5perc_eff = np.array(x_5perc_eff)
    x_10perc_eff = np.array(x_10perc_eff)
    x_30perc_eff = np.array(x_30perc_eff)
    x_50perc_eff = np.array(x_50perc_eff)
    x_90perc_eff = np.array(x_90perc_eff)


    ### remove missing protein expression data (order of these commands important)
    x_orf = x_orf[non_zero_entries]
    x_1perc_eff = x_1perc_eff[non_zero_entries]
    x_5perc_eff = x_5perc_eff[non_zero_entries]
    x_10perc_eff = x_10perc_eff[non_zero_entries]
    x_30perc_eff = x_30perc_eff[non_zero_entries]
    x_50perc_eff = x_50perc_eff[non_zero_entries]
    x_90perc_eff = x_90perc_eff[non_zero_entries]

    ### correlation values for the translation efficiency based CUB index estimates
    corr_vals_eff = []

    ### compute the correalation values stemming from the efficiency-based CUB estimates, for all -1 PRF efficiencies
    corr_vals_eff.append(np.corrcoef(x_1perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_5perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_10perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_30perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_50perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_90perc_eff, log10_y)[0][1])

    plt.plot([0.01, 0.05, 0.1, 0.3, 0.5, 0.9], corr_vals_eff, label = r'$I_c, \pi = $' + str(e) )


### correlation values for the expected-value based CUB index estimates

### define -1 PRF efficiencies that we will look at
p = [0.01, 0.05, 0.1, 0.3, 0.5, 1]

corr_vals_expec = []


for pval in p:

    ### add the correlations coefficient as text
    #log10_y = np.log10(y)

    corr_vals_expec.append(np.corrcoef((1 - p_from_protexp(y, pval)) * x_orf + p_from_protexp(y, pval) * x_prf, log10_y)[0][1])

### plot the results
plt.plot(p, corr_vals_expec, label = r'$I_a$')

### adapt plotting parameters

### set axis limits such that data fall well within them
ax.set_ylim(0, 0.6)

### set axis labels
plt.xlabel('-1 PRF efficiency')
plt.ylabel('Correlation between CAI and log10 of protein expression')


### add legend
plt.legend()


### add panel label
ax.text(-0.1, 0.6*1.075, 'A', fontsize = 20)

########################################################################################################################
#### Figure 5 B
########################################################################################################################

def p_from_protexp(prot_expr, base = 0.1): #, -0.046):

    ### set -1 PRF efficiency to zero at start
    #p = 0.0
    prot_expr = np.array(prot_expr)

    ### p decreases with the inverse of the logarithm of the protein expression
    p = base/(np.log10(prot_expr))

    ### no modification, p is independent from protein expression
    #p = base

    return p

ax = fig.add_subplot(122)


####### expression vs bias indices

dev_meas = 0.75

### declare variables
y = []
x_prf = []

### run a first round, not to repeat below
for i in range(0, len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_prf.append(data_set[i]["CAI_PRF"])

### convert to np.array
y = np.array(y)
x_prf = np.array(x_prf)

### remove non-zero entries
non_zero_entries = [i for i in range(0, len(y)) if y[i] > 0.0]
x_prf = x_prf[non_zero_entries]
y = y[non_zero_entries]

### compute log_10 of protein expression
log10_y = np.log10(y)


### choose all possible values of parameter pi
m_pi = [0, 10, 100, 1000, 10000]

### loop through possible costs of NMD pathway (measured in nucleotide lengths) m_pi
for e in m_pi:

    ### declare variables
    x_orf = []
    x_1perc_eff = []
    x_5perc_eff = []
    x_10perc_eff = []
    x_30perc_eff = []
    x_50perc_eff = []
    x_90perc_eff = []



    for i in range(0,len(data_set)):
        orf_cai = data_set[i]["CAI_ORF"]
        x_orf.append(orf_cai)

        ### ratio of slippery site + cost to gene length
        len_ratio = float(data_set[i]["length_slipseq"] + e)/float(data_set[i]["length_whseq"])

        ### expression level
        prot_expr = data_set[i]["protein_expression"]

        ### fix the -1 PRF efficiency
        p = 0.01
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0/(1.0 + len_ratio*(eff_p/(1-eff_p)))
        x_1perc_eff.append(m_fac*orf_cai)

        p = 0.05
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_5perc_eff.append(m_fac*orf_cai)

        p = 0.1
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_10perc_eff.append(m_fac*orf_cai)

        p = 0.3
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_30perc_eff.append(m_fac*orf_cai)

        p = 0.5
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_50perc_eff.append(m_fac*orf_cai)

        p = 0.9
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_90perc_eff.append(m_fac*orf_cai)

    ### convert to array
    x_orf = np.array(x_orf)
    x_1perc_eff = np.array(x_1perc_eff)
    x_5perc_eff = np.array(x_5perc_eff)
    x_10perc_eff = np.array(x_10perc_eff)
    x_30perc_eff = np.array(x_30perc_eff)
    x_50perc_eff = np.array(x_50perc_eff)
    x_90perc_eff = np.array(x_90perc_eff)


    ### remove missing protein expression data (order of these commands important)
    x_orf = x_orf[non_zero_entries]
    x_1perc_eff = x_1perc_eff[non_zero_entries]
    x_5perc_eff = x_5perc_eff[non_zero_entries]
    x_10perc_eff = x_10perc_eff[non_zero_entries]
    x_30perc_eff = x_30perc_eff[non_zero_entries]
    x_50perc_eff = x_50perc_eff[non_zero_entries]
    x_90perc_eff = x_90perc_eff[non_zero_entries]

    ### correlation values for the translation efficiency based CUB index estimates
    corr_vals_eff = []

    ### compute the correalation values stemming from the efficiency-based CUB estimates, for all -1 PRF efficiencies
    corr_vals_eff.append(np.corrcoef(x_1perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_5perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_10perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_30perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_50perc_eff, log10_y)[0][1])
    corr_vals_eff.append(np.corrcoef(x_90perc_eff, log10_y)[0][1])

    plt.plot([0.01, 0.05, 0.1, 0.3, 0.5, 0.9], corr_vals_eff, label = r'$I_c, \pi = $' + str(e) )


### correlation values for the expected-value based CUB index estimates

### define all the -1 PRF efficinciy we are going to look at
p = [0.01, 0.05, 0.1, 0.3, 0.5, 1]

corr_vals_expec = []

### loop through all of th em
for pval in p:

    corr_vals_expec.append(np.corrcoef((1 - p_from_protexp(y, pval)) * x_orf + p_from_protexp(y, pval) * x_prf, log10_y)[0][1])

### plot the results
plt.plot(p, corr_vals_expec, label = r'$I_a$')


### adapt plotting parameters

### add axis labels
plt.xlabel('baseline -1 PRF efficiency')
plt.ylabel('Correlation between CAI and log10 of protein expression')


### add elgend
plt.legend()

### set axis limits such that data fall well within them
ax.set_ylim(0.45, 0.62)

### add panel label
ax.text(-0.1, 0.45 + (0.62-0.45)*1.075, 'B', fontsize = 20)


### save figure
plt.savefig("figs/Fig5.pdf")


plt.clf()
plt.close()







#######################################################################################################################
################ FIGURE S1:  ################
########################################################################################################################


########################################################################################################################
###  RELATIVE FRAMESHIFTED POST-SLIPPERY SEQUENCE (X_2) LENGTH (LAMBDA) VERSUS REL SLIPPERY SITE POSITION
########################################################################################################################

x_slip_norm = []
x_slip_position = []
x_accession = []
y_lambda_rellength = []
y_0 = []
y_1 = []
y_2 = []


### to see where the information is stored
#test_key = 'YKL029C_309'
#indices[test_key]

### retrieve all the necessary information
for key in indices:
    x_slip_norm.append(indices[key]["slip_start"] / float(indices[key]["length_whseq"]))
    x_slip_position.append(indices[key]["slip_start"])
    y_lambda_rellength.append(indices[key]["length_prfseq"]/ float(indices[key]["length_whseq"]))
    x_accession.append(indices[key]["accession"])
    y_0.append(indices[key]["CAI_0"])
    y_1.append(indices[key]["CAI_1"])
    y_2.append(indices[key]["CAI_2"])



#### test whether makes sense
#plt.scatter(x_slip_norm, y_lambda_rellength, label = 'gene', color = 'blue', edgecolor = 'k')


## Adding an average to the plot
max_x = 1 ### maximum normalized slippery site position
min_x = 0 ### minimum normalized slippery site position
x_maxdiff =  max_x - min_x
n_bins = 20.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_avg_rel_lambda = []
y_avg_0 = []
y_avg_1 = []
y_avg_2 = []

for b in range(1, len(bins)):

    ### for ORF data
    ### these are all the indices of x_orf that fall within the boundaries of the bins
    ind_within_bin = [z for z in range(0,len(x_slip_norm)) if x_slip_norm[z] >= bins[b-1] and x_slip_norm[z] < bins[b]]

    ### here we retrieve all the corresponding y values corresponding to these x's.
    all_snorm_within_bin_rel_lambda = [y_lambda_rellength[z] for z in ind_within_bin]
    all_snorm_within_bin_0 = [y_0[z] for z in ind_within_bin]
    all_snorm_within_bin_1 = [y_1[z] for z in ind_within_bin]
    all_snorm_within_bin_2 = [y_2[z] for z in ind_within_bin]

    ### we average the y values witihin the bin
    y_avg_rel_lambda.append(np.mean(all_snorm_within_bin_rel_lambda))
    y_avg_0.append(np.mean(all_snorm_within_bin_0))
    y_avg_1.append(np.mean(all_snorm_within_bin_1))
    y_avg_2.append(np.mean(all_snorm_within_bin_2))



mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append((bins[z-1] + (bins[z] - bins[z-1])/2.0))


### plot the binned averages
plt.plot(mid_bins, y_avg_rel_lambda, "bo-", label = 'Average relative frameshifted sequence length', linewidth = 2)

plt.ylim(0.0,0.1)
plt.xlim(0,1)

plt.legend(loc = 'upper right')
plt.ylabel('relative frameshifted sequence length')
plt.xlabel('relative slippery site position within gene')

plt.savefig('figs/FigS1.pdf')

plt.clf()




#######################################################################################################################
################ FIGURE S2: HISTOGRAM OF DISTRIBUTION OF SLIPPERY SITES ACROSS GENES FOR ALL FOUND SLIPPERY SITES ################
########################################################################################################################


y_slip_norm = []

for key in indices.keys():
    y_slip_norm.append(indices[key]["slip_start"]/float(indices[key]["length_whseq"]))


## Adding an average to the plot
max_x = 1 ### maximum protein expression at around 10^7
min_x = 0 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 6.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)


########################################################################################################################
############### histogram of distribution of relative slippery site positions
########################################################################################################################


fig = plt.figure(figsize = [6,6]) ### width, height in inches

ax = fig.add_subplot(111)
sns.distplot(a = y_slip_norm, bins = bins, hist=True, kde = False, rug= False )
ax.set_xlim(0,1)


#plt.hist(y_slip_norm, color = 'blue', label = 'frequency', histtype='step', stacked=True, fill=False, linewidth = 2)
plt.xlabel('relative slippery site position')
plt.ylabel('frequency')

### run statistical analysis to determine whether different from zero

### the mean
rel_slip_pos_mean = np.mean(y_slip_norm)
### standard error
rel_slip_pos_std = np.std(y_slip_norm)
### error around the mean
err_slip_pos_mean = 1.96*rel_slip_pos_std/np.sqrt(len(y_slip_norm))

ax.text(0.25, 250, "mean: " + str(round(rel_slip_pos_mean, 4)) + " +/-" + str(round(err_slip_pos_mean, 4)), fontsize = 10)

plt.savefig('figs/FigS2.pdf')
plt.clf()
plt.close()


#######################################################################################################################
################ FIGURE S3: TESTING -1 PRF COST HYPOTHESIS WITH PAXDB DATA SET  ################
########################################################################################################################


### choose the data set that should be plotted

#data_set = ghaemm_data
#data_set = vanderhaar_data
data_set = paxdb_data

#data_slippery = ghaemm_data_slip
#data_slippery = vanderhaar_data_slip
data_slippery = paxdb_data_slip

### Fig S3A

### protein expression
x = []
### number of slippery sites per gene per unit length of gene
y_slip = []

### fill lists by looping through data set
for i in range(0,len(data_slippery)):
    nmb_sites = data_slippery[i]["nmb_slip_sites"]
    length_gene = data_slippery[i]["length_gene"]
    if nmb_sites > 0:
        prot_exp = data_slippery[i]["protein_expression"]
        #x.append(prot_exp)

        mrna_level = data_slippery[i]["mrna_level"]

        if mrna_level > 0.0 and prot_exp > 0.0:
            x.append(float(prot_exp)/float(mrna_level))
        else:
            x.append(prot_exp)

        y_slip.append(float(nmb_sites)/float(length_gene))


### compute the average in bins of equal distance

max_x = 5 ### maximum protein expression at around 10^7
min_x = -2 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 10.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_avg = []

for b in range(1, len(bins)):

    ind_within_bin = [z for z in range(0,len(x)) if x[z] >= 10**bins[b-1] and x[z] < 10**bins[b]]

    all_within_bin = [y_slip[z] for z in ind_within_bin]

    y_avg.append(np.mean(all_within_bin))


mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append(10**(bins[z-1] + (bins[z] - bins[z-1])/2.0) )


######### pyplot ###########################################
####### plot with matplotlib ################
### https://matplotlib.org/users/pyplot_tutorial.html
### http://pbpython.com/effective-matplotlib.html


### style

fig = plt.figure(figsize = [15,6]) ### width, height in inches

### means there will be 1 times 3 panels (1 row, 3 columns),
### and we are concerning ourselves with the first panel
### the counting is 'clockwise' along the rows.
ax = fig.add_subplot(131)

fig.subplots_adjust(top = 0.85)

### scatterplot
plt.plot(x, y_slip, linestyle='', marker='o', markersize = 3, alpha = 0.3)

ax.set_ylim(0,0.01)
#ax.set_xlim(10,10**7)
ax.text(10**min_x, 0.0105, 'A', fontsize = 20)

ax.set_xscale('log')
#plt.legend(loc = 'upper right')


ax.set_ylabel('number of slippery sites per nt per gene')

if mrna_level > 0.0:
    ax.set_xlabel('protein expression [molecules/cell]/mRNA level [molecules/cell]')
else:
    ax.set_xlabel('protein abundance')

### binned averages
plt.plot(mid_bins, y_avg, "ro-", label = 'mean number of slippery sites', linewidth = 2)


### fit a linear regression through the data

### remove all data points where protein expression was 0
y_slip = [y_slip[i] for i in range(0, len(x)) if x[i] > 0.0]
x = [x[i] for i in range(0, len(x)) if x[i] > 0.0]

y_slip = np.array(y_slip)
x = np.array(x)
log10_x = np.log10(x)
m_s,b_s = np.polyfit(log10_x, y_slip, 1)
plt.plot([10**min_x,10**max_x], [min_x*m_s + b_s, max_x*m_s + b_s], linewidth = 3, color = "blue")


### compute the p-value for the null that the regression coefficient differs from 0


### add a text with the value of the regression coefficient to the panels
ax.text(10**2, 0.008, "slope:" + str(round(m_s, 5)), fontsize = 10)


### alternative way to compute regression with statsmodels package
X = add_constant(log10_x)
model = linear_model.OLS(y_slip, X)
results = model.fit()
#print(results.params) -> gives same number as other regression
#print(results.pvalues) # [  2.00489220e-02   4.16826428e-06]
#print(results.t_test)

print results.pvalues[1]

### add panel label text
ax.text(10**2, 0.007, "p-val:" + format_e(results.pvalues[1]), fontsize = 10)

#######################################################################################################################
################ FIGURE S3B: SLIPPERY SITE RELATIVE WITHIN-GENE DISTRIBUTION ################
########################################################################################################################

x_orf = []
x_0_100 = []
x_prf = []
y_slip_position = []
y_slip_norm = []
y_expression = []
y_wh_len = []

for i in range(0,len(data_set)):
    y_slip_position.append(data_set[i]["slip_start"])
    y_slip_norm.append(data_set[i]["slip_start"]/float(data_set[i]["length_whseq"]))
    y_expression.append(data_set[i]["protein_expression"])
    x_orf.append(data_set[i]["CAI_ORF"])
    x_0_100.append(data_set[i]["CAI_0_noramp100"])
    x_prf.append(data_set[i]["CAI_PRF"])
    y_wh_len.append(data_set[i]["length_whseq"])

## Adding an average to the plot
max_x = 1 ### maximum protein expression at around 10^7
min_x = 0 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 6.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_spos_avg_orf = []
y_snorm_avg_orf = []

y_spos_avg_prf = []
y_snorm_avg_prf = []


### loop through all bins
for b in range(1, len(bins)):

    ### for ORF data
    ind_within_bin_orf = [z for z in range(0,len(x_orf)) if x_orf[z] >= bins[b-1] and x_orf[z] < bins[b]]
    all_spos_within_bin_orf = [y_slip_position[z] for z in ind_within_bin_orf]
    all_snorm_within_bin_orf = [y_slip_norm[z] for z in ind_within_bin_orf]

    y_spos_avg_orf.append(np.mean(all_spos_within_bin_orf))
    y_snorm_avg_orf.append(np.mean(all_snorm_within_bin_orf))


    ### for PRF data
    ind_within_bin_prf = [z for z in range(0,len(x_prf)) if x_prf[z] >= bins[b-1] and x_prf[z] < bins[b]]
    all_spos_within_bin_prf = [y_slip_position[z] for z in ind_within_bin_prf]
    all_snorm_within_bin_prf = [y_slip_norm[z] for z in ind_within_bin_prf]

    y_spos_avg_prf.append(np.mean(all_spos_within_bin_prf))
    y_snorm_avg_prf.append(np.mean(all_snorm_within_bin_prf))



mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append((bins[z-1] + (bins[z] - bins[z-1])/2.0))



########################################################################################################################
############### histogram of distribution of relative slippery site positions
########################################################################################################################


ax = fig.add_subplot(132)
sns.distplot(a = y_slip_norm, bins = bins, hist=True, kde = False, rug= False )
ax.set_xlim(0,1)


### set axis labels
plt.xlabel('relative slippery site position')
plt.ylabel('frequency')

### run statistical analysis to determine whether different from one half

### the mean
rel_slip_pos_mean = np.mean(y_slip_norm)
### standard error
rel_slip_pos_std = np.std(y_slip_norm)
### error around the mean
err_slip_pos_mean = 1.96*rel_slip_pos_std/np.sqrt(len(y_slip_norm))

ax.text(0.25, 250, "mean: " + str(round(rel_slip_pos_mean, 4)) + " +/-" + str(round(err_slip_pos_mean, 4)), fontsize = 10)
#ax.text(0.25, 800, "error around mean (+/-): " + str(round(err_slip_pos_mean, 5)), fontsize = 10)


### panel description
ax.text(-0.1, 3200*1.05, 'B', fontsize = 20)


#######################################################################################################################
################ FIGURE S3C: SLIPPERY SITE RELATIVE WITHIN-GENE DISTRIBUTION VS PROTEIN EXPRESSION ################
########################################################################################################################


########################################################################################################################
###  REL SLIPPERY SITE POSITION VS EXPRESSION
########################################################################################################################

y_slip_norm = []
x_expression = []

for i in range(0,len(data_set)):
    y_slip_norm.append(data_set[i]["slip_start"]/float(data_set[i]["length_whseq"]))
    x_expression.append(data_set[i]["protein_expression"])

#### compute averages as well
max_x = 6 ### maximum protein expression at around 10^7
min_x = -2 ### minimum protein expression at around 10^1
x_maxdiff =  max_x - min_x
n_bins = 20.0
bins = np.arange(min_x, max_x + x_maxdiff/n_bins, x_maxdiff/n_bins)

y_slip_norm_avg = []
y_slip_norm_avg_stderr = []

for b in range(1, len(bins)):

    ind_within_bin = [z for z in range(0,len(x_expression)) if x_expression[z] >= 10**bins[b-1] and x_expression[z] < 10**bins[b]]
    all_within_bin = [y_slip_norm[z] for z in ind_within_bin]
    y_slip_norm_avg.append(np.mean(all_within_bin))
    y_slip_norm_avg_stderr.append(1.96*np.std(all_within_bin)/np.sqrt(len(all_within_bin)))


mid_bins = []
for z in range(1, len(bins)):
    mid_bins.append(10**(bins[z-1] + (bins[z] - bins[z-1])/2.0) )


### plot
ax = fig.add_subplot(133)

#plt.scatter(x_expression, y_slip_norm)
plt.plot(x_expression, y_slip_norm, linestyle='', marker='o', markersize = 2, alpha = 0.3 )

#plt.plot(mid_bins, y_slip_norm_avg, 'ro-', linewidth = 2)
plt.errorbar(mid_bins, y_slip_norm_avg, y_slip_norm_avg_stderr, color = 'red', label = 'mean length', linewidth = 2)
plt.plot([10**min_x,10**max_x], [0.5,0.5], 'k', linewidth = 2)

### visualize entries that are non-relevant for analysis due to low power
plt.errorbar(mid_bins[-6::1], y_slip_norm_avg[-6::1], y_slip_norm_avg_stderr[-6::1], color = 'violet', label = 'mean length', linewidth = 2)


#### statistical analysis

### fit a linear regression through the data
### convert to arrays

y_slip_norm = [y_slip_norm[i] for i in range(0, len(x_expression)) if x_expression[i] > 0.0]
x_expression = [x_expression[i] for i in range(0, len(x_expression)) if x_expression[i] > 0.0]


y_slip_norm = np.array(y_slip_norm)
x_expression = np.array(x_expression)
log10_x_expr = np.log10(x_expression)
m_s,b_s = np.polyfit(log10_x_expr, y_slip_norm, 1)

### add the regression line
#plt.plot(x_expression, log10_x_expr*m_s + b_s, linewidth = 1, color = "blue")
plt.plot([10**min_x,10**max_x], [min_x*m_s + b_s, max_x*m_s + b_s], linewidth = 3, color = "blue")

### compute the p-value for the null that the regression coefficient differs from 0

### add a text with the value of the regression coefficient to the panels
ax.text(10**3, 0.2, "slope:" + str(round(m_s, 5)), fontsize = 10)

### alternative way to compute regression with statsmodels package
X = add_constant(log10_x_expr)
model = linear_model.OLS(y_slip_norm, X)
results = model.fit()

### results.params
#print str(results.pvalues[1])

ax.text(10**3, 0.1, "p-val: " + format_e(results.pvalues[1]), fontsize = 10)
#### end of statistical analysis


### set axis limits
plt.ylim(0,1)
plt.xlim(10**min_x,10**max_x)

#ax = plt.gca()
ax.set_xscale('log')

### add panel text
ax.text(10**min_x, 1.05, 'C', fontsize = 20)


plt.xlabel('protein abundance')
plt.ylabel('relative position of slippery site within gene')

plt.savefig('figs/FigS3.pdf')
plt.clf()
plt.close()


### reset to van der Haar data set
#data_set = ghaemm_data
data_set = vanderhaar_data
#data_set = paxdb_data

#data_slippery = ghaemm_data_slip
data_slippery = vanderhaar_data_slip
#data_slippery = paxdb_data_slip

#######################################################################################################################
################ FIGURE S4: FREQUENCY DISTRIBUTION OF CORRECTED AND UNCORRECTED CAI VALUES FOR I_a ESTIMATOR  ################
########################################################################################################################

### more information on plotting histograms etc. here: http://matplotlib.org/api/pyplot_api.html
### and here: https://matplotlib.org/users/pyplot_tutorial.html

### plot log deviations

### set a deviation measure
dev_meas = 0.075
dev_amplifier = 6.5
dev_amp_y = 5


### plot figure
plt.figure(1, figsize= (6,6))

### choosing the right color composition for our figure
### Converting a RGB color tuple to a six digit code

m_cols = []
nmb_cols_used = 2

### set colors
for i in range(0, nmb_cols_used):
    m_precol = colorbrewer.Set1[3][i] #ex_dict[3][0]
    m_color = '#%02x%02x%02x' % m_precol
    m_cols.append(m_color)



####### CAI  ##########################

#plt.subplot(1)

x = []
y = []

for key in indices:
    x.append(indices[key]["CAI_ORF"])
    y.append(indices[key]["CAI_PRF"])


### maximum CAI values
max_x = 1
min_x = 0
x_maxdiff =  max_x - min_x


### convert into arrays to make understandable to matplotlib-function hist
x_a = np.array(x)
y_a = np.array(y)
#m_hist_x = numpy.histogram(x, bins = 20)

### set bins and labels
n_bins = 25.0
m_bins = np.arange(min_x,max_x, x_maxdiff/n_bins)
m_labels =["CAI of ORF", "CAI of -1 PRF mRNA"]

xy_a = [x_a, y_a]
for i in range(0, nmb_cols_used):
    plt.hist(xy_a[i], bins = m_bins, color = m_cols[i], label = m_labels[i], histtype='step', stacked=True, fill=False, linewidth = 2)

### add a legend
plt.legend()

### get plot axis limits
axes = plt.gca()
y_lim = axes.get_ylim()
y_maxdiff = y_lim[1] - y_lim[0]


##### perform kolmogorov smirnov test
ks_p = stats.ks_2samp(x, y)[1]  ### the second entry constitutes the p-value of the two sample KS-test
ks_p = round(ks_p, 6)


### add KS test outcome to the figure (lowest precision is 2e-16
plt.text(min_x + x_maxdiff*(1-dev_amplifier*dev_meas), y_lim[0] + y_maxdiff*(1 - dev_amp_y*dev_meas), 'KS pval < 2e-16')# + str(ks_p))


### add title (name of the index)
plt.title("Codon Adaptation Index")

### add labels
plt.xlabel("CAI")
plt.ylabel("frequency")

### save figure
plt.savefig('figs/FigS4.pdf')

plt.clf()



#######################################################################################################################
################ FIGURE S5: SPEARMAN CORRELATION OF THE ESTIMATOR VALUES TO LOG10(PROT_EXPRESSION) VS -1 PRF EFFICIENCY  ################
########################################################################################################################

def p_from_protexp(prot_expr, base = 0.1): #, -0.046):

    ### set -1 PRF efficiency to zero at start
    #p = 0.0
    #prot_expr = np.array(prot_expr)

    ### p decreases with the inverse of the logarithm of the protein expression
    #p = base/(np.log10(prot_expr))

    ### no modification, p is independent from protein expression
    p = base

    return p


#### start plotting ##############

plt.rcParams.update(plt.rcParamsDefault)
#plt.style.available

plt.style.use('bmh')

### figure
fig = plt.figure(figsize = [12,5]) ### width, height in inches

ax = fig.add_subplot(121)
#fig = plt.figure()


####### expression vs bias indices

dev_meas = 0.75

### declare variables
y = []
x_prf = []

### run a first round, not to repeat below
for i in range(0, len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_prf.append(data_set[i]["CAI_PRF"])

### convert to np.array
y = np.array(y)
x_prf = np.array(x_prf)

### remove non-zero entries
non_zero_entries = [i for i in range(0, len(y)) if y[i] > 0.0]
x_prf = x_prf[non_zero_entries]
y = y[non_zero_entries]

### compute log_10 of protein expression
log10_y = np.log10(y)


### choose all possible values of parameter pi
m_pi = [0, 10, 100, 1000, 10000]

### loop through possible costs of NMD pathway (measured in nucleotide lengths) m_pi
for e in m_pi:

    ### declare variables
    x_orf = []
    x_1perc_eff = []
    x_5perc_eff = []
    x_10perc_eff = []
    x_30perc_eff = []
    x_50perc_eff = []
    x_90perc_eff = []



    for i in range(0,len(data_set)):
        orf_cai = data_set[i]["CAI_ORF"]
        x_orf.append(orf_cai)

        ### ratio of slippery site + cost to gene length
        len_ratio = float(data_set[i]["length_slipseq"] + e)/float(data_set[i]["length_whseq"])

        ### expression level
        prot_expr = data_set[i]["protein_expression"]

        ### fix the -1 PRF efficiency
        p = 0.01
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0/(1.0 + len_ratio*(eff_p/(1-eff_p)))
        x_1perc_eff.append(m_fac*orf_cai)

        p = 0.05
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_5perc_eff.append(m_fac*orf_cai)

        p = 0.1
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_10perc_eff.append(m_fac*orf_cai)

        p = 0.3
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_30perc_eff.append(m_fac*orf_cai)

        p = 0.5
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_50perc_eff.append(m_fac*orf_cai)

        p = 0.9
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_90perc_eff.append(m_fac*orf_cai)

    ### convert to array
    x_orf = np.array(x_orf)
    x_1perc_eff = np.array(x_1perc_eff)
    x_5perc_eff = np.array(x_5perc_eff)
    x_10perc_eff = np.array(x_10perc_eff)
    x_30perc_eff = np.array(x_30perc_eff)
    x_50perc_eff = np.array(x_50perc_eff)
    x_90perc_eff = np.array(x_90perc_eff)


    ### remove missing protein expression data (order of these commands important)
    x_orf = x_orf[non_zero_entries]
    x_1perc_eff = x_1perc_eff[non_zero_entries]
    x_5perc_eff = x_5perc_eff[non_zero_entries]
    x_10perc_eff = x_10perc_eff[non_zero_entries]
    x_30perc_eff = x_30perc_eff[non_zero_entries]
    x_50perc_eff = x_50perc_eff[non_zero_entries]
    x_90perc_eff = x_90perc_eff[non_zero_entries]

    ### correlation values for the translation efficiency based CUB index estimates


    corr_vals_eff = []

    ### compute the correalation values stemming from the efficiency-based CUB estimates, for all -1 PRF efficiencies
    corr_vals_eff.append(stats.spearmanr(x_1perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_5perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_10perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_30perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_50perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_90perc_eff, log10_y)[0])

    plt.plot([0.01, 0.05, 0.1, 0.3, 0.5, 0.9], corr_vals_eff, label = r'$I_c, \pi = $' + str(e) )


### correlation values for the expected-value based CUB index estimates

### define the -1 PRF efficiencies
p = [0.01, 0.05, 0.1, 0.3, 0.5, 1]

corr_vals_expec = []

for pval in p:

    ### add the correlations coefficient as text
    #log10_y = np.log10(y)

    corr_vals_expec.append(stats.spearmanr((1 - p_from_protexp(y, pval)) * x_orf + p_from_protexp(y, pval) * x_prf, log10_y)[0])

### plot the results
plt.plot(p, corr_vals_expec, label = r'$I_a$')


### adapt plotting parameters

### set axis labels and legend
plt.xlabel('-1 PRF efficiency')
plt.ylabel('Spearman rank correlation of CAI to log10(protein expr)')

plt.legend()

### set title
plt.title('constant -1 PRF efficiency')


### add panel label
ax.text(-0.1, 0.47, 'A', fontsize = 20)


#######################################################################################################################
################ FIGURE S5B: CORRELATIONS BETWEEN CAI AND PROTEIN EXPRESSION FOR DECREASING -1 PRF EFFICIENCY WITH PROTEIN EXPRESSION ################



### relationship between p and protein expression
def p_from_protexp(prot_expr, base = 0.1): #, -0.046):

    ### set -1 PRF efficiency to zero at start
    #p = 0.0
    prot_expr = np.array(prot_expr)

    ### p decreases with the inverse of the logarithm of the protein expression
    p = base/(np.log10(prot_expr))

    ### no modification, p is independent from protein expression
    #p = base

    return p


#### start plotting ##############

### set styles
plt.rcParams.update(plt.rcParamsDefault)

plt.style.use('bmh')

### add subplot
ax = fig.add_subplot(122)

####### expression vs bias indices
dev_meas = 0.75

### declare variables
y = []
x_prf = []

### run a first round, not to repeat below
for i in range(0, len(data_set)):
    y.append(data_set[i]["protein_expression"])
    x_prf.append(data_set[i]["CAI_PRF"])

### convert to np.array
y = np.array(y)
x_prf = np.array(x_prf)

### remove non-zero entries
non_zero_entries = [i for i in range(0, len(y)) if y[i] > 0.0]
x_prf = x_prf[non_zero_entries]
y = y[non_zero_entries]

### compute log_10 of protein expression
log10_y = np.log10(y)


### choose all possible values of parameter pi
m_pi = [0, 10, 100, 1000, 10000]

### loop through possible costs of NMD pathway (measured in nucleotide lengths) m_pi
for e in m_pi:

    ### declare variables
    x_orf = []
    x_1perc_eff = []
    x_5perc_eff = []
    x_10perc_eff = []
    x_30perc_eff = []
    x_50perc_eff = []
    x_90perc_eff = []

    for i in range(0,len(data_set)):
        orf_cai = data_set[i]["CAI_ORF"]
        x_orf.append(orf_cai)

        ### ratio of slippery site + cost to gene length
        len_ratio = float(data_set[i]["length_slipseq"] + e)/float(data_set[i]["length_whseq"])

        ### expression level
        prot_expr = data_set[i]["protein_expression"]

        ### fix the -1 PRF efficiency
        p = 0.01
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0/(1.0 + len_ratio*(eff_p/(1-eff_p)))
        x_1perc_eff.append(m_fac*orf_cai)

        p = 0.05
        eff_p = p_from_protexp(prot_expr, base = p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_5perc_eff.append(m_fac*orf_cai)

        p = 0.1
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_10perc_eff.append(m_fac*orf_cai)

        p = 0.3
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_30perc_eff.append(m_fac*orf_cai)

        p = 0.5
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_50perc_eff.append(m_fac*orf_cai)

        p = 0.9
        eff_p = p_from_protexp(prot_expr, base=p)
        m_fac = 1.0 / (1.0 + len_ratio * (eff_p / (1 - eff_p)))
        x_90perc_eff.append(m_fac*orf_cai)

    ### convert to array
    x_orf = np.array(x_orf)
    x_1perc_eff = np.array(x_1perc_eff)
    x_5perc_eff = np.array(x_5perc_eff)
    x_10perc_eff = np.array(x_10perc_eff)
    x_30perc_eff = np.array(x_30perc_eff)
    x_50perc_eff = np.array(x_50perc_eff)
    x_90perc_eff = np.array(x_90perc_eff)


    ### remove missing protein expression data (order of these commands important)
    x_orf = x_orf[non_zero_entries]
    x_1perc_eff = x_1perc_eff[non_zero_entries]
    x_5perc_eff = x_5perc_eff[non_zero_entries]
    x_10perc_eff = x_10perc_eff[non_zero_entries]
    x_30perc_eff = x_30perc_eff[non_zero_entries]
    x_50perc_eff = x_50perc_eff[non_zero_entries]
    x_90perc_eff = x_90perc_eff[non_zero_entries]


    ### correlation values for the translation efficiency based CUB index estimates

    corr_vals_eff = []

    ### compute the correalation values stemming from the efficiency-based CUB estimates, for all -1 PRF efficiencies
    corr_vals_eff.append(stats.spearmanr(x_1perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_5perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_10perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_30perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_50perc_eff, log10_y)[0])
    corr_vals_eff.append(stats.spearmanr(x_90perc_eff, log10_y)[0])

    plt.plot([0.01, 0.05, 0.1, 0.3, 0.5, 0.9], corr_vals_eff, label = r'$I_c, \pi = $' + str(e) )


### correlation values for the expected-value based CUB index estimates

### define -1 PRF efficiencies
p = [0.01, 0.05, 0.1, 0.3, 0.5, 1]

corr_vals_expec = []

### loop through all values of p
for pval in p:

    ### add the correlations coefficient as text
    #log10_y = np.log10(y)

    corr_vals_expec.append(stats.spearmanr((1 - p_from_protexp(y, pval)) * x_orf + p_from_protexp(y, pval) * x_prf, log10_y)[0])

### plot the results
plt.plot(p, corr_vals_expec, label = r'$I_a$')


### adapt plotting parameters

### add axis labels
plt.xlabel('baseline -1 PRF efficiency')
plt.ylabel('Spearman rank correlation of CAI to log10(protein expr)')


### add legend
plt.legend()


plt.title('-1 PRF efficiency depends on protein expression')


### add panel label
ax.text(-0.1, 0.56, 'B', fontsize = 20)



### save figure
plt.savefig("figs/FigS5.pdf")

plt.clf()
plt.close()





